"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var emu_1 = require("./emu");
var util_1 = require("./util");
var disasm6502_1 = require("./cpu/disasm6502");
var disasmz80_1 = require("./cpu/disasmz80");
;
;
var DebugSymbols = /** @class */ (function () {
    function DebugSymbols(symbolmap) {
        this.symbolmap = symbolmap;
        this.addr2symbol = util_1.invertMap(symbolmap);
        // TODO: shouldn't be necc.
        if (!this.addr2symbol[0x0])
            this.addr2symbol[0x0] = '__START__'; // needed for ...
        this.addr2symbol[0x10000] = '__END__'; // ... dump memory to work
    }
    return DebugSymbols;
}());
exports.DebugSymbols = DebugSymbols;
// for composite breakpoints w/ single debug function
var BreakpointList = /** @class */ (function () {
    function BreakpointList() {
        this.id2bp = {};
    }
    BreakpointList.prototype.getDebugCondition = function () {
        var _this = this;
        if (Object.keys(this.id2bp).length == 0) {
            return null; // no breakpoints
        }
        else {
            // evaluate all breakpoints
            return function () {
                var result = false;
                for (var id in _this.id2bp)
                    if (_this.id2bp[id].cond())
                        result = true;
                return result;
            };
        }
    };
    return BreakpointList;
}());
exports.BreakpointList = BreakpointList;
/////
var BasePlatform = /** @class */ (function () {
    function BasePlatform() {
        this.recorder = null;
    }
    BasePlatform.prototype.setRecorder = function (recorder) {
        this.recorder = recorder;
    };
    BasePlatform.prototype.updateRecorder = function () {
        // are we recording and do we need to save a frame?
        if (this.recorder && this.isRunning() && this.recorder.frameRequested()) {
            this.recorder.recordFrame(this.saveState());
        }
    };
    return BasePlatform;
}());
exports.BasePlatform = BasePlatform;
var BaseDebugPlatform = /** @class */ (function (_super) {
    __extends(BaseDebugPlatform, _super);
    function BaseDebugPlatform() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.debugSavedState = null;
        _this.debugBreakState = null;
        _this.debugTargetClock = 0;
        _this.debugClock = 0;
        _this.breakpoints = new BreakpointList();
        return _this;
    }
    BaseDebugPlatform.prototype.setBreakpoint = function (id, cond) {
        if (cond) {
            this.breakpoints.id2bp[id] = { cond: cond };
            this.restartDebugging();
        }
        else {
            this.clearBreakpoint(id);
        }
    };
    BaseDebugPlatform.prototype.clearBreakpoint = function (id) {
        delete this.breakpoints.id2bp[id];
    };
    BaseDebugPlatform.prototype.getDebugCallback = function () {
        return this.breakpoints.getDebugCondition();
    };
    BaseDebugPlatform.prototype.setupDebug = function (callback) {
        this.onBreakpointHit = callback;
    };
    BaseDebugPlatform.prototype.clearDebug = function () {
        this.debugSavedState = null;
        this.debugBreakState = null;
        this.debugTargetClock = -1;
        this.debugClock = 0;
        this.onBreakpointHit = null;
        this.clearBreakpoint('debug');
    };
    BaseDebugPlatform.prototype.setDebugCondition = function (debugCond) {
        this.setBreakpoint('debug', debugCond);
    };
    BaseDebugPlatform.prototype.restartDebugging = function () {
        if (this.debugSavedState) {
            this.loadState(this.debugSavedState);
        }
        else {
            this.debugSavedState = this.saveState();
        }
        this.debugClock = 0;
        this.debugCallback = this.getDebugCallback();
        this.debugBreakState = null;
        this.resume();
    };
    BaseDebugPlatform.prototype.preFrame = function () {
    };
    BaseDebugPlatform.prototype.postFrame = function () {
    };
    BaseDebugPlatform.prototype.pollControls = function () {
    };
    BaseDebugPlatform.prototype.nextFrame = function (novideo) {
        this.pollControls();
        this.updateRecorder();
        this.preFrame();
        this.advance(novideo);
        this.postFrame();
    };
    return BaseDebugPlatform;
}(BasePlatform));
exports.BaseDebugPlatform = BaseDebugPlatform;
////// 6502
function getToolForFilename_6502(fn) {
    if (fn.endsWith(".pla"))
        return "plasm";
    if (fn.endsWith(".c"))
        return "cc65";
    if (fn.endsWith(".h"))
        return "cc65";
    if (fn.endsWith(".s"))
        return "ca65";
    if (fn.endsWith(".ca65"))
        return "ca65";
    if (fn.endsWith(".dasm"))
        return "dasm";
    if (fn.endsWith(".acme"))
        return "acme";
    return "dasm"; // .a
}
exports.getToolForFilename_6502 = getToolForFilename_6502;
// TODO: can merge w/ Z80?
var Base6502Platform = /** @class */ (function (_super) {
    __extends(Base6502Platform, _super);
    function Base6502Platform() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        // some platforms store their PC one byte before or after the first opcode
        // so we correct when saving and loading from state
        _this.debugPCDelta = -1;
        _this.getToolForFilename = getToolForFilename_6502;
        return _this;
    }
    Base6502Platform.prototype.fixPC = function (c) { c.PC = (c.PC + this.debugPCDelta) & 0xffff; return c; };
    Base6502Platform.prototype.unfixPC = function (c) { c.PC = (c.PC - this.debugPCDelta) & 0xffff; return c; };
    Base6502Platform.prototype.evalDebugCondition = function () {
        if (this.debugCallback && !this.debugBreakState) {
            this.debugCallback();
        }
    };
    Base6502Platform.prototype.postFrame = function () {
        if (this.debugCallback) {
            if (this.debugBreakState) {
                // reload debug state at end of frame after breakpoint
                this.loadState(this.debugBreakState);
            }
            else {
                // save state every frame and rewind debug clocks
                this.debugSavedState = this.saveState();
                this.debugTargetClock -= this.debugClock;
                this.debugClock = 0;
            }
        }
    };
    Base6502Platform.prototype.breakpointHit = function (targetClock) {
        this.debugTargetClock = targetClock;
        this.debugBreakState = this.saveState();
        console.log("Breakpoint at clk", this.debugClock, "PC", this.debugBreakState.c.PC.toString(16));
        this.pause();
        if (this.onBreakpointHit) {
            this.onBreakpointHit(this.debugBreakState);
        }
    };
    Base6502Platform.prototype.runEval = function (evalfunc) {
        var _this = this;
        this.setDebugCondition(function () {
            if (_this.debugClock++ > _this.debugTargetClock) {
                var cpuState = _this.getCPUState();
                if (evalfunc(cpuState)) {
                    _this.breakpointHit(_this.debugClock - 1);
                    return true;
                }
                else {
                    return false;
                }
            }
        });
    };
    Base6502Platform.prototype.runToFrameClock = function (clock) {
        var _this = this;
        this.restartDebugging();
        this.debugTargetClock = clock;
        this.setDebugCondition(function () {
            if (_this.debugClock++ > _this.debugTargetClock) {
                _this.breakpointHit(_this.debugClock - 1);
                return true;
            }
        });
    };
    Base6502Platform.prototype.step = function () {
        var _this = this;
        var previousPC = -1;
        this.setDebugCondition(function () {
            //console.log(this.debugClock, this.debugTargetClock, this.getCPUState().PC, this.getCPUState());
            if (_this.debugClock++ >= _this.debugTargetClock) {
                var thisState = _this.getCPUState();
                if (previousPC < 0) {
                    previousPC = thisState.PC;
                }
                else {
                    // doesn't work w/ endless loops
                    if (thisState.PC != previousPC && thisState.T == 0) {
                        _this.breakpointHit(_this.debugClock - 1);
                        return true;
                    }
                }
            }
            return false;
        });
    };
    Base6502Platform.prototype.stepBack = function () {
        var _this = this;
        var prevState;
        var prevClock;
        this.setDebugCondition(function () {
            if (_this.debugClock++ >= _this.debugTargetClock && prevState) {
                _this.loadState(prevState);
                _this.breakpointHit(prevClock - 1);
                return true;
            }
            else if (_this.debugClock > _this.debugTargetClock - 10 && _this.debugClock <= _this.debugTargetClock + _this.debugPCDelta) { // TODO: why this works?
                if (_this.getCPUState().T == 0) {
                    prevState = _this.saveState();
                    prevClock = _this.debugClock;
                }
            }
            return false;
        });
    };
    Base6502Platform.prototype.newCPU = function (membus) {
        var cpu = new jt.M6502();
        cpu.connectBus(membus);
        return cpu;
    };
    Base6502Platform.prototype.getOpcodeMetadata = function (opcode, offset) {
        return getOpcodeMetadata_6502(opcode, offset);
    };
    Base6502Platform.prototype.getOriginPC = function () {
        return (this.readAddress(0xfffc) | (this.readAddress(0xfffd) << 8)) & 0xffff;
    };
    Base6502Platform.prototype.runUntilReturn = function () {
        var depth = 1;
        this.runEval(function (c) {
            if (depth <= 0 && c.T == 0)
                return true;
            if (c.o == 0x20)
                depth++;
            else if (c.o == 0x60 || c.o == 0x40)
                --depth;
            return false;
        });
    };
    Base6502Platform.prototype.disassemble = function (pc, read) {
        return disasm6502_1.disassemble6502(pc, read(pc), read(pc + 1), read(pc + 2));
    };
    Base6502Platform.prototype.getDefaultExtension = function () { return ".a"; };
    ;
    Base6502Platform.prototype.getDebugCategories = function () {
        return ['CPU', 'ZPRAM', 'Stack'];
    };
    Base6502Platform.prototype.getDebugInfo = function (category, state) {
        switch (category) {
            case 'CPU': return cpuStateToLongString_6502(state.c);
            case 'ZPRAM': return emu_1.dumpRAM(state.b, 0x0, 0x100);
            case 'Stack': return dumpStackToString(this, state.b, 0x100, 0x1ff, 0x100 + state.c.SP, 0x20);
        }
    };
    return Base6502Platform;
}(BaseDebugPlatform));
exports.Base6502Platform = Base6502Platform;
function cpuStateToLongString_6502(c) {
    function decodeFlags(c) {
        var s = "";
        s += c.N ? " N" : " -";
        s += c.V ? " V" : " -";
        s += c.D ? " D" : " -";
        s += c.Z ? " Z" : " -";
        s += c.C ? " C" : " -";
        s += c.I ? " I" : " -";
        return s;
    }
    return "PC " + util_1.hex(c.PC, 4) + "  " + decodeFlags(c) + "\n"
        + " A " + util_1.hex(c.A) + "     " + (c.R ? "" : "BUSY") + "\n"
        + " X " + util_1.hex(c.X) + "\n"
        + " Y " + util_1.hex(c.Y) + "     " + "SP " + util_1.hex(c.SP) + "\n";
}
exports.cpuStateToLongString_6502 = cpuStateToLongString_6502;
var OPMETA_6502 = {
    cycletime: [
        7, 6, 0, 8, 3, 3, 5, 5, 3, 2, 2, 2, 4, 4, 6, 6, 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 0, 7, 4, 4, 7, 7, 6, 6, 0, 8, 3, 3, 5, 5, 4, 2, 2, 2, 4, 4, 6, 6, 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 0, 7, 4, 4, 7, 7, 6, 6, 0, 8, 3, 3, 5, 5, 3, 2, 2, 2, 3, 4, 6, 6, 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 0, 7, 4, 4, 7, 7, 6, 6, 0, 8, 3, 3, 5, 5, 4, 2, 2, 2, 5, 4, 6, 6, 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 0, 7, 4, 4, 7, 7, 0, 6, 0, 6, 3, 3, 3, 3, 2, 0, 2, 0, 4, 4, 4, 4, 2, 6, 0, 0, 4, 4, 4, 4, 2, 5, 2, 0, 0, 5, 0, 0, 2, 6, 2, 6, 3, 3, 3, 3, 2, 2, 2, 0, 4, 4, 4, 4, 2, 5, 0, 5, 4, 4, 4, 4, 2, 4, 2, 0, 4, 4, 4, 4, 2, 6, 0, 8, 3, 3, 5, 5, 2, 2, 2, 2, 4, 4, 3, 6, 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 0, 7, 4, 4, 7, 7, 2, 6, 0, 8, 3, 3, 5, 5, 2, 2, 2, 0, 4, 4, 6, 6, 2, 5, 0, 8, 4, 4, 6, 6, 2, 4, 0, 7, 4, 4, 7, 7
    ],
    extracycles: [
        0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1
    ],
    insnlengths: [
        1, 2, 0, 2, 2, 2, 2, 2, 1, 2, 1, 2, 3, 3, 3, 3, 2, 2, 0, 2, 2, 2, 2, 2, 1, 3, 0, 3, 3, 3, 3, 3, 3, 2, 0, 2, 2, 2, 2, 2, 1, 2, 1, 2, 3, 3, 3, 3, 2, 2, 0, 2, 2, 2, 2, 2, 1, 3, 0, 3, 3, 3, 3, 3, 1, 2, 0, 2, 2, 2, 2, 2, 1, 2, 1, 2, 3, 3, 3, 3, 2, 2, 0, 2, 2, 2, 2, 2, 1, 3, 0, 3, 3, 3, 3, 3, 1, 2, 0, 2, 2, 2, 2, 2, 1, 2, 1, 2, 3, 3, 3, 3, 2, 2, 0, 2, 2, 2, 2, 2, 1, 3, 0, 3, 3, 3, 3, 3, 0, 2, 0, 2, 2, 2, 2, 2, 1, 0, 1, 0, 3, 3, 3, 3, 2, 2, 0, 0, 2, 2, 2, 3, 1, 3, 1, 0, 0, 3, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 1, 2, 1, 0, 3, 3, 3, 3, 2, 2, 0, 2, 2, 2, 2, 2, 1, 3, 1, 0, 3, 3, 3, 3, 2, 2, 0, 2, 2, 2, 2, 2, 1, 2, 1, 2, 3, 3, 3, 3, 2, 2, 0, 2, 2, 2, 2, 2, 1, 3, 0, 3, 3, 3, 3, 3, 2, 2, 0, 2, 2, 2, 2, 2, 1, 2, 1, 0, 3, 3, 3, 3, 2, 2, 0, 2, 2, 2, 2, 2, 1, 3, 0, 3, 3, 3, 3, 3
    ],
    validinsns: [
        1, 2, 0, 0, 0, 2, 2, 0, 1, 2, 1, 0, 0, 3, 3, 0, 2, 2, 0, 0, 0, 2, 2, 0, 1, 3, 0, 0, 0, 3, 3, 0, 3, 2, 0, 0, 2, 2, 2, 0, 1, 2, 1, 0, 3, 3, 3, 0, 2, 2, 0, 0, 0, 2, 2, 0, 1, 3, 0, 0, 0, 3, 3, 0, 1, 2, 0, 0, 0, 2, 2, 0, 1, 2, 1, 0, 3, 3, 3, 0, 2, 2, 0, 0, 0, 2, 2, 0, 1, 3, 0, 0, 0, 3, 3, 0, 1, 2, 0, 0, 0, 2, 2, 0, 1, 2, 1, 0, 3, 3, 3, 0, 2, 2, 0, 0, 0, 2, 2, 0, 1, 3, 0, 0, 0, 3, 3, 0, 0, 2, 0, 0, 2, 2, 2, 0, 1, 0, 1, 0, 3, 3, 3, 0, 2, 2, 0, 0, 2, 2, 2, 0, 1, 3, 1, 0, 0, 3, 0, 0, 2, 2, 2, 0, 2, 2, 2, 0, 1, 2, 1, 0, 3, 3, 3, 0, 2, 2, 0, 0, 2, 2, 2, 0, 1, 3, 1, 0, 3, 3, 3, 0, 2, 2, 0, 0, 2, 2, 2, 0, 1, 2, 1, 0, 3, 3, 3, 0, 2, 2, 0, 0, 0, 2, 2, 0, 1, 3, 0, 0, 0, 3, 3, 0, 2, 2, 0, 0, 2, 2, 2, 0, 1, 2, 1, 0, 3, 3, 3, 0, 2, 2, 0, 0, 0, 2, 2, 0, 1, 3, 0, 0, 0, 3, 3, 0
    ],
};
function getOpcodeMetadata_6502(opcode, address) {
    // TODO: more intelligent maximum cycles
    // TODO: must always be new object, b/c we might modify it
    return {
        opcode: opcode,
        minCycles: OPMETA_6502.cycletime[opcode],
        maxCycles: OPMETA_6502.cycletime[opcode] + OPMETA_6502.extracycles[opcode],
        insnlength: OPMETA_6502.insnlengths[opcode]
    };
}
exports.getOpcodeMetadata_6502 = getOpcodeMetadata_6502;
////// Z80
function cpuStateToLongString_Z80(c) {
    function decodeFlags(flags) {
        return util_1.printFlags(flags, ["S", "Z", , "H", , "V", "N", "C"], true);
    }
    return "PC " + util_1.hex(c.PC, 4) + "  " + decodeFlags(c.AF) + "\n"
        + "SP " + util_1.hex(c.SP, 4) + "  IR " + util_1.hex(c.IR, 4) + "\n"
        + "IX " + util_1.hex(c.IX, 4) + "  IY " + util_1.hex(c.IY, 4) + "\n"
        + "AF " + util_1.hex(c.AF, 4) + "  BC " + util_1.hex(c.BC, 4) + "\n"
        + "DE " + util_1.hex(c.DE, 4) + "  HL " + util_1.hex(c.HL, 4) + "\n";
}
exports.cpuStateToLongString_Z80 = cpuStateToLongString_Z80;
function BusProbe(bus) {
    var active = false;
    var callback;
    this.activate = function (_callback) {
        active = true;
        callback = _callback;
    };
    this.deactivate = function () {
        active = false;
        callback = null;
    };
    this.read = function (a) {
        if (active) {
            callback(a);
        }
        return bus.read(a);
    };
    this.write = function (a, v) {
        if (active) {
            callback(a, v);
        }
        bus.write(a, v);
    };
    this.contend = function (addr, cyc) {
        return bus.contend(addr, cyc);
    };
    this.isContended = function (addr) {
        return bus.isContended(addr);
    };
}
exports.BusProbe = BusProbe;
var BaseZ80Platform = /** @class */ (function (_super) {
    __extends(BaseZ80Platform, _super);
    function BaseZ80Platform() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.getToolForFilename = getToolForFilename_z80;
        return _this;
    }
    BaseZ80Platform.prototype.newCPU = function (membus, iobus, z80opts) {
        this.probe = new BusProbe(membus);
        this._cpu = buildZ80(z80opts || {})({
            display: {},
            memory: this.probe,
            ioBus: iobus
        });
        return this._cpu;
    };
    BaseZ80Platform.prototype.getProbe = function () { return this.probe; }; // TODO?
    BaseZ80Platform.prototype.getPC = function () { return this._cpu.getPC(); };
    BaseZ80Platform.prototype.getSP = function () { return this._cpu.getSP(); };
    // TODO: refactor other parts into here
    BaseZ80Platform.prototype.runCPU = function (cpu, cycles) {
        this._cpu = cpu; // TODO?
        if (this.wasBreakpointHit())
            return 0;
        var debugCond = this.getDebugCallback();
        var targetTstates = cpu.getTstates() + cycles;
        try {
            if (debugCond) { // || trace) {
                while (cpu.getTstates() < targetTstates) {
                    if (debugCond && debugCond()) {
                        debugCond = null;
                        break;
                    }
                    cpu.runFrame(cpu.getTstates() + 1);
                }
            }
            else {
                cpu.runFrame(targetTstates);
            }
        }
        catch (e) {
            // TODO: show alert w/ error msg
            console.log(e);
            this.breakpointHit(cpu.getTstates());
        }
        return cpu.getTstates() - targetTstates;
    };
    BaseZ80Platform.prototype.requestInterrupt = function (cpu, data) {
        if (!this.wasBreakpointHit())
            cpu.requestInterrupt(data);
    };
    BaseZ80Platform.prototype.postFrame = function () {
        if (this.debugCallback) {
            if (this.debugBreakState) {
                // if breakpoint, reload debug state after frame
                this.loadState(this.debugBreakState);
            }
            else {
                // reset debug target clocks
                this.debugSavedState = this.saveState();
                if (this.debugTargetClock > 0)
                    this.debugTargetClock -= this.debugSavedState.c.T;
                this.debugSavedState.c.T = 0;
                this.loadState(this.debugSavedState);
            }
        }
    };
    BaseZ80Platform.prototype.breakpointHit = function (targetClock) {
        this.debugTargetClock = targetClock;
        this.debugBreakState = this.saveState();
        console.log("Breakpoint at clk", this.debugBreakState.c.T, "PC", this.debugBreakState.c.PC.toString(16));
        this.pause();
        if (this.onBreakpointHit) {
            this.onBreakpointHit(this.debugBreakState);
        }
    };
    BaseZ80Platform.prototype.wasBreakpointHit = function () {
        return this.debugBreakState != null;
    };
    // TODO: lower bound of clock value
    BaseZ80Platform.prototype.step = function () {
        var _this = this;
        this.setDebugCondition(function () {
            var cpuState = _this.getCPUState();
            if (cpuState.T > _this.debugTargetClock) {
                _this.breakpointHit(cpuState.T);
                return true;
            }
            return false;
        });
    };
    BaseZ80Platform.prototype.stepBack = function () {
        var _this = this;
        var prevState;
        var prevClock;
        this.setDebugCondition(function () {
            var cpuState = _this.getCPUState();
            var debugClock = cpuState.T;
            if (debugClock >= _this.debugTargetClock && prevState) {
                _this.loadState(prevState);
                _this.breakpointHit(prevClock);
                return true;
            }
            else if (debugClock > _this.debugTargetClock - 20 && debugClock < _this.debugTargetClock) {
                prevState = _this.saveState();
                prevClock = debugClock;
            }
            return false;
        });
    };
    BaseZ80Platform.prototype.runEval = function (evalfunc) {
        var _this = this;
        this.setDebugCondition(function () {
            var cpuState = _this.getCPUState();
            if (cpuState.T > _this.debugTargetClock) {
                if (evalfunc(cpuState)) {
                    _this.breakpointHit(cpuState.T);
                    return true;
                }
            }
            return false;
        });
    };
    BaseZ80Platform.prototype.runUntilReturn = function () {
        var _this = this;
        var depth = 1;
        this.runEval(function (c) {
            if (depth <= 0)
                return true;
            var op = _this.readAddress(c.PC);
            if (op == 0xcd) // CALL
                depth++;
            else if (op == 0xc0 || op == 0xc8 || op == 0xc9 || op == 0xd0) // RET (TODO?)
                --depth;
            return false;
        });
    };
    BaseZ80Platform.prototype.runToVsync = function () {
        this.runEval(function (c) { return c['intp']; });
    };
    BaseZ80Platform.prototype.getDefaultExtension = function () { return ".c"; };
    ;
    // TODO: Z80 opcode metadata
    //this.getOpcodeMetadata = function() { }
    BaseZ80Platform.prototype.getDebugCategories = function () {
        return ['CPU', 'Stack'];
    };
    BaseZ80Platform.prototype.getDebugInfo = function (category, state) {
        switch (category) {
            case 'CPU': return cpuStateToLongString_Z80(state.c);
            case 'Stack': {
                var sp = (state.c.SP - 1) & 0xffff;
                var start = sp & 0xff00;
                var end = start + 0xff;
                if (sp == 0)
                    sp = 0x10000;
                console.log(sp, start, end);
                return dumpStackToString(this, [], start, end, sp, 0xcd);
            }
        }
    };
    BaseZ80Platform.prototype.disassemble = function (pc, read) {
        return disasmz80_1.disassembleZ80(pc, read(pc), read(pc + 1), read(pc + 2), read(pc + 3));
    };
    return BaseZ80Platform;
}(BaseDebugPlatform));
exports.BaseZ80Platform = BaseZ80Platform;
function getToolForFilename_z80(fn) {
    if (fn.endsWith(".c"))
        return "sdcc";
    if (fn.endsWith(".h"))
        return "sdcc";
    if (fn.endsWith(".s"))
        return "sdasz80";
    if (fn.endsWith(".ns"))
        return "naken";
    if (fn.endsWith(".scc"))
        return "sccz80";
    if (fn.endsWith(".z"))
        return "zmac";
    return "zmac";
}
exports.getToolForFilename_z80 = getToolForFilename_z80;
////// 6809
function cpuStateToLongString_6809(c) {
    function decodeFlags(flags) {
        return util_1.printFlags(flags, ["E", "F", "H", "I", "N", "Z", "V", "C"], true);
    }
    return "PC " + util_1.hex(c.PC, 4) + "  " + decodeFlags(c.CC) + "\n"
        + "SP " + util_1.hex(c.SP, 4) + "\n"
        + " A " + util_1.hex(c.A, 2) + "\n"
        + " B " + util_1.hex(c.B, 2) + "\n"
        + " X " + util_1.hex(c.X, 4) + "\n"
        + " Y " + util_1.hex(c.Y, 4) + "\n"
        + " U " + util_1.hex(c.U, 4) + "\n";
}
exports.cpuStateToLongString_6809 = cpuStateToLongString_6809;
var Base6809Platform = /** @class */ (function (_super) {
    __extends(Base6809Platform, _super);
    function Base6809Platform() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        //this.getOpcodeMetadata = function() { }
        _this.getToolForFilename = function () { return "xasm6809"; };
        return _this;
    }
    Base6809Platform.prototype.newCPU = function (membus) {
        var cpu = new CPU6809();
        cpu.init(membus.write, membus.read, 0);
        return cpu;
    };
    Base6809Platform.prototype.getPC = function () { return this._cpu.PC; };
    Base6809Platform.prototype.getSP = function () { return this._cpu.SP; };
    Base6809Platform.prototype.runUntilReturn = function () {
        var _this = this;
        var depth = 1;
        this.runEval(function (c) {
            if (depth <= 0)
                return true;
            var op = _this.readAddress(c.PC);
            // TODO: 6809 opcodes
            if (op == 0x9d || op == 0xad || op == 0xbd) // CALL
                depth++;
            else if (op == 0x3b || op == 0x39) // RET
                --depth;
            return false;
        });
    };
    Base6809Platform.prototype.cpuStateToLongString = function (c) {
        return cpuStateToLongString_6809(c);
    };
    Base6809Platform.prototype.disassemble = function (pc, read) {
        // TODO: don't create new CPU
        return new CPU6809().disasm(read(pc), read(pc + 1), read(pc + 2), read(pc + 3), read(pc + 4), pc);
    };
    Base6809Platform.prototype.getDefaultExtension = function () { return ".asm"; };
    ;
    Base6809Platform.prototype.getDebugCategories = function () {
        return ['CPU', 'Stack'];
    };
    Base6809Platform.prototype.getDebugInfo = function (category, state) {
        switch (category) {
            case 'CPU': return cpuStateToLongString_6809(state.c);
            default: return _super.prototype.getDebugInfo.call(this, category, state);
        }
    };
    return Base6809Platform;
}(BaseZ80Platform));
exports.Base6809Platform = Base6809Platform;
var BaseMAMEPlatform = /** @class */ (function () {
    function BaseMAMEPlatform(mainElement) {
        this.loaded = false;
        this.preinitted = false;
        this.started = false;
        this.running = false;
        this.console_vars = {};
        this.initluavars = false;
        this.mainElement = mainElement;
    }
    BaseMAMEPlatform.prototype.luareset = function () {
        this.console_vars = {};
    };
    // http://docs.mamedev.org/techspecs/luaengine.html
    BaseMAMEPlatform.prototype.luacall = function (s) {
        this.console_varname = null;
        //Module.ccall('_Z13js_lua_stringPKc', 'void', ['string'], [s+""]);
        if (!this.js_lua_string)
            this.js_lua_string = Module.cwrap('_Z13js_lua_stringPKc', 'void', ['string']);
        this.js_lua_string(s || "");
    };
    BaseMAMEPlatform.prototype.pause = function () {
        if (this.loaded && this.running) {
            this.luacall('emu.pause()');
            this.running = false;
        }
    };
    BaseMAMEPlatform.prototype.resume = function () {
        if (this.loaded && !this.running) { // TODO
            this.luacall('emu.unpause()');
            this.running = true;
        }
    };
    BaseMAMEPlatform.prototype.reset = function () {
        if (this.loaded) {
            this.luacall('manager:machine():soft_reset()');
            this.running = true;
            this.initluavars = false;
        }
    };
    BaseMAMEPlatform.prototype.isRunning = function () {
        return this.running;
    };
    BaseMAMEPlatform.prototype.bufferConsoleOutput = function (s) {
        if (!s)
            return;
        if (s.startsWith(">>>")) {
            this.console_varname = s.length > 3 ? s.slice(3) : null;
            if (this.console_varname)
                this.console_vars[this.console_varname] = [];
        }
        else if (this.console_varname) {
            this.console_vars[this.console_varname].push(s);
            if (this.console_varname == 'debug_stopped') {
                var debugSaveState = this.preserveState();
                this.pause();
                if (this.onBreakpointHit) {
                    this.onBreakpointHit(debugSaveState);
                }
            }
        }
        else {
            console.log(s);
        }
    };
    BaseMAMEPlatform.prototype.startModule = function (mainElement, opts) {
        var _this = this;
        this.started = true;
        var romfn = this.romfn = this.romfn || opts.romfn;
        var romdata = this.romdata = this.romdata || opts.romdata || new emu_1.RAM(opts.romsize).mem;
        // create canvas
        var video = this.video = new emu_1.RasterVideo(this.mainElement, opts.width, opts.height);
        video.create();
        $(video.canvas).attr('id', 'canvas');
        // load asm.js module
        console.log("loading", opts.jsfile);
        var modargs = [opts.driver,
            '-debug',
            '-debugger', 'none',
            '-verbose', '-window', '-nokeepaspect',
            '-resolution', video.canvas.width + 'x' + video.canvas.height
        ];
        if (romfn)
            modargs.push('-cart', romfn);
        window['JSMESS'] = {};
        window['Module'] = {
            arguments: modargs,
            screenIsReadOnly: true,
            print: this.bufferConsoleOutput,
            canvas: video.canvas,
            doNotCaptureKeyboard: true,
            keyboardListeningElement: video.canvas,
            preInit: function () {
                console.log("loading FS");
                ENV.SDL_EMSCRIPTEN_KEYBOARD_ELEMENT = 'canvas';
                if (opts.cfgfile) {
                    FS.mkdir('/cfg');
                    FS.writeFile('/cfg/' + opts.cfgfile, opts.cfgdata, { encoding: 'utf8' });
                }
                if (opts.biosfile) {
                    FS.mkdir('/roms');
                    FS.mkdir('/roms/' + opts.driver);
                    FS.writeFile('/roms/' + opts.biosfile, opts.biosdata, { encoding: 'binary' });
                }
                FS.mkdir('/emulator');
                if (romfn) {
                    FS.writeFile(romfn, romdata, { encoding: 'binary' });
                }
                //FS.writeFile('/debug.ini', 'debugger none\n', {encoding:'utf8'});
                if (opts.preInit) {
                    opts.preInit(self);
                }
                _this.preinitted = true;
            },
            preRun: [
                function () {
                    $(video.canvas).click(function (e) {
                        video.canvas.focus();
                    });
                    _this.loaded = true;
                    console.log("about to run...");
                }
            ]
        };
        // preload files
        // TODO: ensure loaded
        var fetch_cfg, fetch_lua;
        var fetch_bios = $.Deferred();
        var fetch_wasm = $.Deferred();
        // fetch config file
        if (opts.cfgfile) {
            fetch_cfg = $.get('mame/cfg/' + opts.cfgfile, function (data) {
                opts.cfgdata = data;
                console.log("loaded " + opts.cfgfile);
            }, 'text');
        }
        // fetch BIOS file
        if (opts.biosfile) {
            var oReq1 = new XMLHttpRequest();
            oReq1.open("GET", 'mame/roms/' + opts.biosfile, true);
            oReq1.responseType = "arraybuffer";
            oReq1.onload = function (oEvent) {
                opts.biosdata = new Uint8Array(oReq1.response);
                console.log("loaded " + opts.biosfile); // + " (" + oEvent.total + " bytes)");
                fetch_bios.resolve();
            };
            oReq1.send();
        }
        else {
            fetch_bios.resolve();
        }
        // load debugger Lua script
        fetch_lua = $.get('mame/debugger.lua', function (data) {
            _this.luadebugscript = data;
            console.log("loaded debugger.lua");
        }, 'text');
        // load WASM
        {
            var oReq2 = new XMLHttpRequest();
            oReq2.open("GET", 'mame/' + opts.jsfile.replace('.js', '.wasm'), true);
            oReq2.responseType = "arraybuffer";
            oReq2.onload = function (oEvent) {
                console.log("loaded WASM file");
                window['Module'].wasmBinary = new Uint8Array(oReq2.response);
                fetch_wasm.resolve();
            };
            oReq2.send();
        }
        // start loading script
        $.when(fetch_lua, fetch_cfg, fetch_bios, fetch_wasm).done(function () {
            var script = document.createElement('script');
            script.src = 'mame/' + opts.jsfile;
            document.getElementsByTagName('head')[0].appendChild(script);
            console.log("created script element");
        });
    };
    BaseMAMEPlatform.prototype.loadROMFile = function (data) {
        this.romdata = data;
        if (this.preinitted && this.romfn) {
            FS.writeFile(this.romfn, data, { encoding: 'binary' });
        }
    };
    BaseMAMEPlatform.prototype.loadRegion = function (region, data) {
        if (this.loaded && data.length > 0) {
            //this.luacall('cart=manager:machine().images["cart"]\nprint(cart:filename())\ncart:load("' + romfn + '")\n');
            var s = 'rgn = manager:machine():memory().regions["' + region + '"]\n';
            //s += 'print(rgn.size)\n';
            for (var i = 0; i < data.length; i += 4) {
                var v = data[i] + (data[i + 1] << 8) + (data[i + 2] << 16) + (data[i + 3] << 24);
                s += 'rgn:write_u32(' + i + ',' + v + ')\n'; // TODO: endian?
            }
            this.luacall(s);
            this.reset();
        }
    };
    BaseMAMEPlatform.prototype.preserveState = function () {
        var state = { c: {} };
        for (var k in this.console_vars) {
            if (k.startsWith("cpu_")) {
                var v = parseInt(this.console_vars[k][0]);
                state.c[k.slice(4)] = v;
            }
        }
        // TODO: memory?
        return state;
    };
    /*
      saveState() {
        this.luareset();
        this.luacall('mamedbg.printstate()');
        return this.preserveState();
      }
    */
    BaseMAMEPlatform.prototype.initlua = function () {
        if (!this.initluavars) {
            this.luacall(this.luadebugscript);
            this.luacall('mamedbg.init()');
            this.initluavars = true;
        }
    };
    BaseMAMEPlatform.prototype.readAddress = function (a) {
        this.initlua();
        this.luacall('print(">>>v"); print(mem:read_u8(' + a + '))');
        return parseInt(this.console_vars.v[0]);
    };
    // DEBUGGING SUPPORT
    BaseMAMEPlatform.prototype.clearDebug = function () {
        this.onBreakpointHit = null;
    };
    BaseMAMEPlatform.prototype.getDebugCallback = function () {
        return this.onBreakpointHit; // TODO?
    };
    BaseMAMEPlatform.prototype.setupDebug = function (callback) {
        if (this.loaded) { // TODO?
            this.initlua();
            this.luareset();
        }
        this.onBreakpointHit = callback;
    };
    BaseMAMEPlatform.prototype.runToPC = function (pc) {
        this.luacall('mamedbg.runTo(' + pc + ')');
        this.resume();
    };
    BaseMAMEPlatform.prototype.runToVsync = function () {
        this.luacall('mamedbg.runToVsync()');
        this.resume();
    };
    BaseMAMEPlatform.prototype.runUntilReturn = function () {
        this.luacall('mamedbg.runUntilReturn()');
        this.resume();
    };
    BaseMAMEPlatform.prototype.step = function () {
        this.luacall('mamedbg.step()');
        this.resume();
    };
    // TODO: other than z80
    BaseMAMEPlatform.prototype.cpuStateToLongString = function (c) {
        if (c.HL)
            return cpuStateToLongString_Z80(c);
        else
            return null; // TODO
    };
    return BaseMAMEPlatform;
}());
exports.BaseMAMEPlatform = BaseMAMEPlatform;
function dumpStackToString(platform, mem, start, end, sp, jsrop) {
    var s = "";
    var nraw = 0;
    //s = dumpRAM(mem.slice(start,start+end+1), start, end-start+1);
    function read(addr) {
        if (addr < mem.length)
            return mem[addr];
        else
            return platform.readAddress(addr);
    }
    while (sp < end) {
        sp++;
        // see if there's a JSR on the stack here
        // TODO: make work with roms and memory maps
        var addr = read(sp) + read(sp + 1) * 256;
        var jsrofs = jsrop == 0x20 ? -2 : -3; // 6502 vs Z80
        var opcode = read(addr + jsrofs); // might be out of bounds
        if (opcode == jsrop) { // JSR
            s += "\n$" + util_1.hex(sp) + ": ";
            s += util_1.hex(addr, 4) + " " + lookupSymbol(platform, addr, true);
            sp++;
            nraw = 0;
        }
        else {
            if (nraw == 0)
                s += "\n$" + util_1.hex(sp) + ": ";
            s += util_1.hex(read(sp)) + " ";
            if (++nraw == 8)
                nraw = 0;
        }
    }
    return s + "\n";
}
exports.dumpStackToString = dumpStackToString;
function lookupSymbol(platform, addr, extra) {
    var start = addr;
    var addr2symbol = platform.debugSymbols && platform.debugSymbols.addr2symbol;
    while (addr2symbol && addr >= 0) {
        var sym = addr2symbol[addr];
        if (sym) { // return first symbol we find
            var sym = addr2symbol[addr];
            return extra ? (sym + " + " + (start - addr)) : sym;
        }
        addr--;
    }
    return "";
}
exports.lookupSymbol = lookupSymbol;
///// Basic Platforms
var BasicZ80ScanlinePlatform = /** @class */ (function (_super) {
    __extends(BasicZ80ScanlinePlatform, _super);
    function BasicZ80ScanlinePlatform(mainElement) {
        var _this = _super.call(this) || this;
        _this.ram = new Uint8Array(0);
        _this.rom = new Uint8Array(0);
        _this.inputs = new Uint8Array(16);
        _this.mainElement = mainElement;
        return _this;
    }
    BasicZ80ScanlinePlatform.prototype.getRasterScanline = function () { return this.currentScanline; };
    BasicZ80ScanlinePlatform.prototype.getKeyboardFunction = function () { return null; };
    BasicZ80ScanlinePlatform.prototype.start = function () {
        this.cpuCyclesPerLine = Math.round(this.cpuFrequency / 60 / this.numTotalScanlines);
        this.ram = this.newRAM();
        this.membus = this.newMembus();
        this.iobus = this.newIOBus();
        this.cpu = this.newCPU(this.membus, this.iobus);
        this.video = new emu_1.RasterVideo(this.mainElement, this.canvasWidth, this.numVisibleScanlines, this.getVideoOptions());
        this.video.create();
        this.pixels = this.video.getFrameData();
        this.poller = emu_1.setKeyboardFromMap(this.video, this.inputs, this.getKeyboardMap(), this.getKeyboardFunction());
        this.timer = new emu_1.AnimationTimer(60, this.nextFrame.bind(this));
    };
    BasicZ80ScanlinePlatform.prototype.readAddress = function (addr) {
        return this.membus.read(addr);
    };
    BasicZ80ScanlinePlatform.prototype.pollControls = function () { this.poller.poll(); };
    BasicZ80ScanlinePlatform.prototype.advance = function (novideo) {
        var extraCycles = 0;
        for (var sl = 0; sl < this.numTotalScanlines; sl++) {
            this.startLineTstates = this.cpu.getTstates();
            this.currentScanline = sl;
            this.startScanline(sl);
            extraCycles = this.runCPU(this.cpu, this.cpuCyclesPerLine - extraCycles); // TODO: HALT opcode?
            this.drawScanline(sl);
        }
        this.video.updateFrame();
    };
    BasicZ80ScanlinePlatform.prototype.loadROM = function (title, data) {
        this.rom = emu_1.padBytes(data, this.defaultROMSize);
        this.reset();
    };
    BasicZ80ScanlinePlatform.prototype.loadState = function (state) {
        this.cpu.loadState(state.c);
        this.ram.set(state.b);
        this.inputs.set(state.in);
    };
    BasicZ80ScanlinePlatform.prototype.saveState = function () {
        return {
            c: this.getCPUState(),
            b: this.ram.slice(0),
            in: this.inputs.slice(0),
        };
    };
    BasicZ80ScanlinePlatform.prototype.loadControlsState = function (state) {
        this.inputs.set(state.in);
    };
    BasicZ80ScanlinePlatform.prototype.saveControlsState = function () {
        return {
            in: this.inputs.slice(0)
        };
    };
    BasicZ80ScanlinePlatform.prototype.getCPUState = function () {
        return this.cpu.saveState();
    };
    BasicZ80ScanlinePlatform.prototype.isRunning = function () {
        return this.timer && this.timer.isRunning();
    };
    BasicZ80ScanlinePlatform.prototype.pause = function () {
        this.timer.stop();
        if (this.audio)
            this.audio.stop();
    };
    BasicZ80ScanlinePlatform.prototype.resume = function () {
        this.timer.start();
        if (this.audio)
            this.audio.start();
    };
    BasicZ80ScanlinePlatform.prototype.reset = function () {
        this.cpu.reset();
        this.cpu.setTstates(0);
    };
    return BasicZ80ScanlinePlatform;
}(BaseZ80Platform));
exports.BasicZ80ScanlinePlatform = BasicZ80ScanlinePlatform;
//# sourceMappingURL=baseplatform.js.map