"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var baseplatform_1 = require("../baseplatform");
var emu_1 = require("../emu");
var util_1 = require("../util");
var audio_1 = require("../audio");
// https://atarihq.com/danb/a7800.shtml
// https://atarihq.com/danb/files/maria_r1.txt
// https://sites.google.com/site/atari7800wiki/
var Atari7800_PRESETS = [
    { id: 'sprites.dasm', name: 'Sprites (ASM)' },
];
var SWCHA = 0;
var SWCHB = 1;
var INPT4 = 2;
var INPT5 = 3;
var Atari7800_KEYCODE_MAP = emu_1.makeKeycodeMap([
    [emu_1.Keys.A, INPT4, -0x80],
    //[Keys.B,        0, 1],
    //[Keys.SELECT,   0, 2],
    //[Keys.START,    0, 3],
    [emu_1.Keys.UP, SWCHA, -0x10],
    [emu_1.Keys.DOWN, SWCHA, -0x20],
    [emu_1.Keys.LEFT, SWCHA, -0x40],
    [emu_1.Keys.RIGHT, SWCHA, -0x80],
    [emu_1.Keys.P2_A, INPT5, -0x80],
    //[Keys.P2_B,      1, 1],
    //[Keys.P2_SELECT, 1, 2],
    //[Keys.P2_START,  1, 3],
    [emu_1.Keys.P2_UP, SWCHA, -0x01],
    [emu_1.Keys.P2_DOWN, SWCHA, -0x02],
    [emu_1.Keys.P2_LEFT, SWCHA, -0x04],
    [emu_1.Keys.P2_RIGHT, SWCHA, -0x08],
]);
// http://www.ataripreservation.org/websites/freddy.offenga/megazine/ISSUE5-PALNTSC.html
// http://7800.8bitdev.org/index.php/7800_Software_Guide#APPENDIX_4:_FRAME_TIMING
var CLK = 3579545;
var cpuFrequency = 1789772;
var linesPerFrame = 262;
var colorClocksPerLine = 454;
var romLength = 0xc000;
// TIA chip
var TIA = /** @class */ (function () {
    function TIA() {
        this.regs = new Uint8Array(0x20);
    }
    TIA.prototype.reset = function () {
        // TODO
    };
    TIA.prototype.read = function (a) {
        return this.regs[a] | 0;
    };
    TIA.prototype.write = function (a, v) {
        this.regs[a] = v;
    };
    TIA.prototype.saveState = function () {
        return {
            regs: this.regs.slice(0)
        };
    };
    TIA.prototype.loadState = function (s) {
        for (var i_1 = 0; i_1 < 32; i_1++)
            this.write(i_1, s.regs[i_1]);
    };
    TIA.stateToLongString = function (state) {
        var s = "";
        s += emu_1.dumpRAM(state.regs, 0, 32);
        return s;
    };
    return TIA;
}());
// MARIA chip
var MARIA = /** @class */ (function () {
    function MARIA() {
        this.regs = new Uint8Array(0x20);
        this.offset = -1;
        this.dll = 0;
        this.dlstart = 0;
        this.dli = false;
        this.h16 = false;
        this.h8 = false;
        this.pixels = new Uint8Array(320);
    }
    MARIA.prototype.reset = function () {
        // TODO
    };
    MARIA.prototype.read = function (a) {
        return this.regs[a] | 0;
    };
    MARIA.prototype.write = function (a, v) {
        this.regs[a] = v;
        //console.log(hex(a), '=', hex(v));
    };
    MARIA.prototype.saveState = function () {
        return {
            regs: this.regs.slice(0),
            offset: this.offset,
            dll: this.dll,
            dlstart: this.dlstart,
            dli: this.dli,
            h16: this.h16,
            h8: this.h8,
        };
    };
    MARIA.prototype.loadState = function (s) {
        for (var i_2 = 0; i_2 < 32; i_2++)
            this.write(i_2, s.regs[i_2]);
        this.offset = s.offset;
        this.dll = s.dll;
        this.dlstart = s.dlstart;
        this.dli = s.dli;
        this.h16 = s.h16;
        this.h8 = s.h8;
    };
    MARIA.prototype.isDMAEnabled = function () {
        return (this.regs[0x1c] & 0x60) == 0x40;
    };
    MARIA.prototype.getDLLStart = function () {
        return (this.regs[0x0c] << 8) + this.regs[0x10];
    };
    MARIA.prototype.getCharBaseAddress = function () {
        return (this.regs[0x14] << 8) + this.offset;
    };
    MARIA.prototype.setVBLANK = function (b) {
        if (b) {
            this.regs[0x08] |= 0x80;
            this.offset = -1;
            this.dll = this.getDLLStart(); // TODO?
        }
        else {
            this.regs[0x08] &= ~0x80;
        }
    };
    MARIA.prototype.readDLLEntry = function (bus) {
        var x = bus.read(this.dll);
        this.offset = (x & 0xf);
        this.dli = (x & 0x80) != 0;
        this.h16 = (x & 0x40) != 0;
        this.h8 = (x & 0x20) != 0;
        this.dlstart = (bus.read(this.dll + 1) << 8) + bus.read(this.dll + 2);
        //console.log(hex(this.dll), this.offset, hex(this.dlstart,4));
        this.dll = (this.dll + 3) & 0xffff; // TODO: can only cross 1 page
    };
    MARIA.prototype.doDMA = function (bus) {
        var cycles = 0;
        this.pixels.fill(this.regs[0x0]);
        if (this.isDMAEnabled()) {
            cycles += 16;
            // time for a new DLL entry?
            if (this.offset < 0) {
                this.readDLLEntry(bus);
                cycles += 8;
            }
            // read the DL
            var dla = this.dlstart;
            do {
                // read DL entry
                var b0 = bus.read(dla);
                var b1 = bus.read(dla + 1);
                if (b1 == 0)
                    break; // end of DL
                var b2 = bus.read(dla + 2);
                var b3 = bus.read(dla + 3);
                var gfxadr = b0 + (((b2 + this.offset) & 0xff) << 8);
                var writemode = 0;
                var indirect = false;
                // extended header?
                if (b1 & 0x40) {
                    var pal = b3 >> 5;
                    var width = 32 - (b3 & 31);
                    var xpos = bus.read(dla + 4);
                    writemode = b1 & 0x80;
                    indirect = (b1 & 0x20) != 0;
                    dla += 5;
                }
                else {
                    // direct mode
                    var xpos = b3;
                    var pal = b1 >> 5;
                    var width = 32 - (b1 & 31);
                    dla += 4;
                }
                xpos *= 2;
                // next DL entry
                if (dla > 0xffff) {
                    break; // TODO: throw exception
                }
                cycles += 4;
                // copy graphics data (direct)
                //console.log(hex(gfxadr,4), xpos, width, pal);
                // TODO
                var readmode = (this.regs[0x1c] & 0x3) + (writemode ? 4 : 0);
                for (var i = 0; i < width; i++) {
                    var data = bus.read(gfxadr + i);
                    switch (readmode) {
                        case 0: // 160 A/B
                            for (var j = 0; j < 4; j++) {
                                var col = (data >> 6) & 3;
                                if (col > 0) {
                                    this.pixels[xpos] = this.pixels[xpos + 1] = this.regs[(pal << 2) + col];
                                }
                                data <<= 2;
                                xpos = (xpos + 2) & 0x1ff;
                            }
                            break;
                        case 2: // 320 B/D
                        case 3: // 320 A/C
                            for (var j = 0; j < 8; j++) {
                                var col = data & 1;
                                if (col > 0) {
                                    this.pixels[xpos] = this.regs[(pal << 2) + col];
                                }
                                data >>= 1;
                                xpos = (xpos + 1) & 0x1ff;
                            }
                            break;
                    }
                    cycles += 3;
                }
            } while (1);
            // decrement offset
            this.offset -= 1;
        }
        return cycles;
    };
    MARIA.prototype.doInterrupt = function () {
        return this.dli && this.offset < 0;
    };
    MARIA.stateToLongString = function (state) {
        var s = "";
        s += emu_1.dumpRAM(state.regs, 0, 32);
        s += "\n   DLL: $" + util_1.hex(state.dll, 4);
        s += "\n    DL: $" + util_1.hex(state.dlstart, 4);
        s += "\nOffset:  " + state.offset;
        s += "\n   DLI?  " + state.dli;
        return s;
    };
    return MARIA;
}());
// Atari 7800
var Atari7800Platform = /** @class */ (function (_super) {
    __extends(Atari7800Platform, _super);
    function Atari7800Platform(mainElement) {
        var _this = _super.call(this) || this;
        _this.inputs = new Uint8Array(8);
        _this.scanline = 0;
        _this.tia = new TIA();
        _this.maria = new MARIA();
        _this.mainElement = mainElement;
        _this.inputs.fill(0xff);
        return _this;
    }
    Atari7800Platform.prototype.getPresets = function () {
        return Atari7800_PRESETS;
    };
    Atari7800Platform.prototype.start = function () {
        var _this = this;
        this.cpu = new jt.M6502();
        this.ram = new Uint8Array(0x2800 - 0x1600);
        this.bios = new Uint8Array(0x1000); // TODO
        // TODO: TIA access wastes a cycle
        this.bus = {
            read: emu_1.newAddressDecoder([
                [0x0008, 0x000d, 0x1f, function (a) { return _this.inputs[a - 4]; }],
                [0x0000, 0x001f, 0x1f, function (a) { return _this.tia.read(a); }],
                [0x0020, 0x003f, 0x1f, function (a) { return _this.maria.read(a); }],
                [0x0040, 0x00ff, 0xffff, function (a) { return _this.ram[a]; }],
                [0x0140, 0x01ff, 0xffff, function (a) { return _this.ram[a]; }],
                [0x0280, 0x0283, 0x3, function (a) { return _this.inputs[a]; }],
                [0x1800, 0x27ff, 0xffff, function (a) { return _this.ram[a - 0x1600]; }],
                [0x4000, 0xffff, 0xffff, function (a) { return _this.rom ? _this.rom[a - 0x4000] : 0; }],
            ]),
            write: emu_1.newAddressDecoder([
                [0x0000, 0x001f, 0x1f, function (a, v) { _this.tia.write(a, v); }],
                [0x0020, 0x003f, 0x1f, function (a, v) { _this.maria.write(a, v); }],
                [0x0040, 0x00ff, 0xffff, function (a, v) { _this.ram[a] = v; }],
                [0x0140, 0x01ff, 0xffff, function (a, v) { _this.ram[a] = v; }],
                [0x1800, 0x27ff, 0xffff, function (a, v) { _this.ram[a - 0x1600] = v; }],
            ]),
        };
        this.cpu.connectBus(this.bus);
        // create video/audio
        this.video = new emu_1.RasterVideo(this.mainElement, 320, 192);
        this.audio = audio_1.newPOKEYAudio(1);
        this.video.create();
        emu_1.setKeyboardFromMap(this.video, this.inputs, Atari7800_KEYCODE_MAP, function (o, key, code, flags) {
            // TODO
        });
        this.timer = new emu_1.AnimationTimer(60, this.nextFrame.bind(this));
    };
    Atari7800Platform.prototype.advance = function (novideo) {
        var idata = this.video.getFrameData();
        var iofs = 0;
        var debugCond = this.getDebugCallback();
        var rgb;
        var mariaClocks = 0;
        // visible lines
        for (var sl = 0; sl < linesPerFrame; sl++) {
            this.maria.setVBLANK(sl >= 192);
            mariaClocks += 454;
            // TODO: 7 cycles at start of line
            if (this.maria.isDMAEnabled()) {
                // do DMA for scanline
                mariaClocks -= this.maria.doDMA(this.bus);
                // copy line to frame buffer
                for (var i = 0; i < 320; i++) {
                    idata[iofs++] = COLORS_RGBA[this.maria.pixels[i]];
                }
                // do interrupt?
                if (this.maria.doInterrupt()) {
                    mariaClocks -= this.cpu.setNMIAndWait() * 4;
                }
            }
            // iterate CPU with free clocks
            while (mariaClocks > 0) {
                mariaClocks -= 4;
                if (debugCond && debugCond()) {
                    debugCond = null;
                    break;
                }
                this.cpu.clockPulse();
            }
        }
        // update video frame
        if (!novideo) {
            this.video.updateFrame();
            // set background/border color
            var bkcol = this.maria.regs[0x0];
            $(this.video.canvas).css('background-color', COLORS_WEB[bkcol]);
        }
    };
    Atari7800Platform.prototype.loadROM = function (title, data) {
        this.rom = emu_1.padBytes(data, romLength, true);
        this.reset();
    };
    Atari7800Platform.prototype.loadBIOS = function (title, data) {
        this.bios = emu_1.padBytes(data, 0x1000);
        this.reset();
    };
    Atari7800Platform.prototype.isRunning = function () {
        return this.timer.isRunning();
    };
    Atari7800Platform.prototype.pause = function () {
        this.timer.stop();
        this.audio.stop();
    };
    Atari7800Platform.prototype.resume = function () {
        this.timer.start();
        this.audio.start();
    };
    Atari7800Platform.prototype.reset = function () {
        this.cpu.reset();
        this.tia.reset();
        this.maria.reset();
        /*
        // execute until out of BIOS
        for (var i=0; i<20000; i++) {
          this.cpu.clockPulse();
          if (this.getCPUState().PC < 0xf000)
            break;
        }
        */
    };
    Atari7800Platform.prototype.readAddress = function (addr) {
        return (addr < 0x1800 ? this.ram[addr] : this.bus.read(addr)) | 0;
    };
    Atari7800Platform.prototype.loadState = function (state) {
        this.unfixPC(state.c);
        this.cpu.loadState(state.c);
        this.fixPC(state.c);
        this.ram.set(state.b);
        this.tia.loadState(state.tia);
        this.maria.loadState(state.maria);
        this.loadControlsState(state);
    };
    Atari7800Platform.prototype.saveState = function () {
        return {
            c: this.getCPUState(),
            b: this.ram.slice(0),
            tia: this.tia.saveState(),
            maria: this.maria.saveState(),
            in: this.inputs.slice(0)
        };
    };
    Atari7800Platform.prototype.loadControlsState = function (state) {
        this.inputs.set(state.in);
    };
    Atari7800Platform.prototype.saveControlsState = function () {
        return {
            in: this.inputs.slice(0)
        };
    };
    Atari7800Platform.prototype.getCPUState = function () {
        return this.fixPC(this.cpu.saveState());
    };
    Atari7800Platform.prototype.getRasterScanline = function () {
        return this.scanline;
    };
    Atari7800Platform.prototype.getDebugCategories = function () {
        return _super.prototype.getDebugCategories.call(this).concat(['TIA', 'MARIA']);
    };
    Atari7800Platform.prototype.getDebugInfo = function (category, state) {
        switch (category) {
            case 'TIA': return TIA.stateToLongString(state.tia);
            case 'MARIA': return MARIA.stateToLongString(state.maria) + "\nScanline: " + this.scanline;
            default: return _super.prototype.getDebugInfo.call(this, category, state);
        }
    };
    return Atari7800Platform;
}(baseplatform_1.Base6502Platform));
///
var ATARI_NTSC_RGB = [
    0x000000,
    0x404040,
    0x6c6c6c,
    0x909090,
    0xb0b0b0,
    0xc8c8c8,
    0xdcdcdc,
    0xf4f4f4,
    0x004444,
    0x106464,
    0x248484,
    0x34a0a0,
    0x40b8b8,
    0x50d0d0,
    0x5ce8e8,
    0x68fcfc,
    0x002870,
    0x144484,
    0x285c98,
    0x3c78ac,
    0x4c8cbc,
    0x5ca0cc,
    0x68b4dc,
    0x78c8ec,
    0x001884,
    0x183498,
    0x3050ac,
    0x4868c0,
    0x5c80d0,
    0x7094e0,
    0x80a8ec,
    0x94bcfc,
    0x000088,
    0x20209c,
    0x3c3cb0,
    0x5858c0,
    0x7070d0,
    0x8888e0,
    0xa0a0ec,
    0xb4b4fc,
    0x5c0078,
    0x74208c,
    0x883ca0,
    0x9c58b0,
    0xb070c0,
    0xc084d0,
    0xd09cdc,
    0xe0b0ec,
    0x780048,
    0x902060,
    0xa43c78,
    0xb8588c,
    0xcc70a0,
    0xdc84b4,
    0xec9cc4,
    0xfcb0d4,
    0x840014,
    0x982030,
    0xac3c4c,
    0xc05868,
    0xd0707c,
    0xe08894,
    0xeca0a8,
    0xfcb4bc,
    0x880000,
    0x9c201c,
    0xb04038,
    0xc05c50,
    0xd07468,
    0xe08c7c,
    0xeca490,
    0xfcb8a4,
    0x7c1800,
    0x90381c,
    0xa85438,
    0xbc7050,
    0xcc8868,
    0xdc9c7c,
    0xecb490,
    0xfcc8a4,
    0x5c2c00,
    0x784c1c,
    0x906838,
    0xac8450,
    0xc09c68,
    0xd4b47c,
    0xe8cc90,
    0xfce0a4,
    0x2c3c00,
    0x485c1c,
    0x647c38,
    0x809c50,
    0x94b468,
    0xacd07c,
    0xc0e490,
    0xd4fca4,
    0x003c00,
    0x205c20,
    0x407c40,
    0x5c9c5c,
    0x74b474,
    0x8cd08c,
    0xa4e4a4,
    0xb8fcb8,
    0x003814,
    0x1c5c34,
    0x387c50,
    0x50986c,
    0x68b484,
    0x7ccc9c,
    0x90e4b4,
    0xa4fcc8,
    0x00302c,
    0x1c504c,
    0x347068,
    0x4c8c84,
    0x64a89c,
    0x78c0b4,
    0x88d4cc,
    0x9cece0,
    0x002844,
    0x184864,
    0x306884,
    0x4484a0,
    0x589cb8,
    0x6cb4d0,
    0x7ccce8,
    0x8ce0fc // FE
];
var COLORS_RGBA = new Uint32Array(256);
var COLORS_WEB = [];
for (var i = 0; i < 256; i++) {
    COLORS_RGBA[i] = ATARI_NTSC_RGB[i >> 1] | 0xff000000;
    COLORS_WEB[i] = "#" + util_1.hex(util_1.rgb2bgr(ATARI_NTSC_RGB[i >> 1]), 6);
}
///
emu_1.PLATFORMS['atari7800'] = Atari7800Platform;
//# sourceMappingURL=atari7800.js.map