"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var baseplatform_1 = require("../baseplatform");
var emu_1 = require("../emu");
// http://www.computerarcheology.com/Arcade/
var MW8080BW_PRESETS = [
    { id: 'gfxtest.c', name: 'Graphics Test' },
    { id: 'shifter.c', name: 'Sprite w/ Bit Shifter' },
    { id: 'game2.c', name: 'Cosmic Impalas' },
];
var SPACEINV_KEYCODE_MAP = emu_1.makeKeycodeMap([
    [emu_1.Keys.A, 1, 0x10],
    [emu_1.Keys.LEFT, 1, 0x20],
    [emu_1.Keys.RIGHT, 1, 0x40],
    [emu_1.Keys.P2_A, 2, 0x10],
    [emu_1.Keys.P2_LEFT, 2, 0x20],
    [emu_1.Keys.P2_RIGHT, 2, 0x40],
    [emu_1.Keys.SELECT, 1, 0x1],
    [emu_1.Keys.START, 1, 0x4],
    [emu_1.Keys.P2_START, 1, 0x2],
]);
var INITIAL_WATCHDOG = 256;
var PIXEL_ON = 0xffeeeeee;
var PIXEL_OFF = 0xff000000;
var Midway8080BWPlatform = /** @class */ (function (_super) {
    __extends(Midway8080BWPlatform, _super);
    function Midway8080BWPlatform() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.cpuFrequency = 1996800; // MHz
        _this.canvasWidth = 256;
        _this.numTotalScanlines = 262;
        _this.numVisibleScanlines = 224;
        _this.defaultROMSize = 0x2000;
        _this.bitshift_offset = 0;
        _this.bitshift_register = 0;
        return _this;
    }
    Midway8080BWPlatform.prototype.getPresets = function () { return MW8080BW_PRESETS; };
    Midway8080BWPlatform.prototype.getKeyboardMap = function () { return SPACEINV_KEYCODE_MAP; };
    Midway8080BWPlatform.prototype.getVideoOptions = function () { return { rotate: -90 }; };
    Midway8080BWPlatform.prototype.newRAM = function () { return new Uint8Array(0x2000); };
    Midway8080BWPlatform.prototype.newMembus = function () {
        var _this = this;
        return {
            read: emu_1.newAddressDecoder([
                [0x0000, 0x1fff, 0x1fff, function (a) { return _this.rom ? _this.rom[a] : 0; }],
                [0x2000, 0x3fff, 0x1fff, function (a) { return _this.ram[a]; }],
            ]),
            write: emu_1.newAddressDecoder([
                [0x2000, 0x23ff, 0x3ff, function (a, v) { _this.ram[a] = v; }],
                [0x2400, 0x3fff, 0x1fff, function (a, v) {
                        _this.ram[a] = v;
                        var ofs = (a - 0x400) << 3;
                        for (var i = 0; i < 8; i++) {
                            _this.pixels[ofs + i] = (v & (1 << i)) ? PIXEL_ON : PIXEL_OFF;
                        }
                        //if (displayPCs) displayPCs[a] = cpu.getPC(); // save program counter
                    }],
            ]),
        };
    };
    Midway8080BWPlatform.prototype.newIOBus = function () {
        var _this = this;
        return {
            read: function (addr) {
                addr &= 0x3;
                //console.log('IO read', hex(addr,4));
                switch (addr) {
                    case 0:
                    case 1:
                    case 2:
                        return _this.inputs[addr];
                    case 3:
                        return (_this.bitshift_register >> (8 - _this.bitshift_offset)) & 0xff;
                }
                return 0;
            },
            write: function (addr, val) {
                addr &= 0x7;
                val &= 0xff;
                //console.log('IO write', hex(addr,4), hex(val,2));
                switch (addr) {
                    case 2:
                        _this.bitshift_offset = val & 0x7;
                        break;
                    case 3:
                    case 5:
                        // TODO: sound
                        break;
                    case 4:
                        _this.bitshift_register = (_this.bitshift_register >> 8) | (val << 8);
                        break;
                    case 6:
                        _this.watchdog_counter = INITIAL_WATCHDOG;
                        break;
                }
            }
        };
    };
    Midway8080BWPlatform.prototype.startScanline = function (sl) {
    };
    Midway8080BWPlatform.prototype.drawScanline = function (sl) {
        // at end of scanline
        if (sl == 95)
            this.cpu.requestInterrupt(0x8); // RST $8
        else if (sl == 223)
            this.cpu.requestInterrupt(0x10); // RST $10
    };
    Midway8080BWPlatform.prototype.advance = function (novideo) {
        _super.prototype.advance.call(this, novideo);
        if (this.watchdog_counter-- <= 0) {
            console.log("WATCHDOG FIRED"); // TODO: alert on video
            this.reset();
        }
    };
    Midway8080BWPlatform.prototype.loadState = function (state) {
        _super.prototype.loadState.call(this, state);
        this.bitshift_register = state.bsr;
        this.bitshift_offset = state.bso;
        this.watchdog_counter = state.wdc;
    };
    Midway8080BWPlatform.prototype.saveState = function () {
        var state = _super.prototype.saveState.call(this);
        state.bsr = this.bitshift_register;
        state.bso = this.bitshift_offset;
        state.wdc = this.watchdog_counter;
        return state;
    };
    Midway8080BWPlatform.prototype.reset = function () {
        _super.prototype.reset.call(this);
        this.watchdog_counter = INITIAL_WATCHDOG;
    };
    return Midway8080BWPlatform;
}(baseplatform_1.BasicZ80ScanlinePlatform));
/*
        $(video.canvas).click(function(e) {
            var x = Math.floor(e.offsetX * video.canvas.width / $(video.canvas).width());
            var y = Math.floor(e.offsetY * video.canvas.height / $(video.canvas).height());
            var addr = (x>>3) + (y*32) + 0x400;
      if (displayPCs) console.log(x, y, hex(addr,4), "PC", hex(displayPCs[addr],4));
        });
  */
emu_1.PLATFORMS['mw8080bw'] = Midway8080BWPlatform;
//# sourceMappingURL=mw8080bw.js.map