"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var $ = require("jquery");
var baseplatform_1 = require("./baseplatform");
var util_1 = require("./util");
var ui_1 = require("./ui");
var recorder_1 = require("./recorder");
var pixed = require("./pixed/pixeleditor");
;
// helper function for editor
function jumpToLine(ed, i) {
    var t = ed.charCoords({ line: i, ch: 0 }, "local").top;
    var middleHeight = ed.getScrollerElement().offsetHeight / 2;
    ed.scrollTo(null, t - middleHeight - 5);
}
function createTextSpan(text, className) {
    var span = document.createElement("span");
    span.setAttribute("class", className);
    span.appendChild(document.createTextNode(text));
    return span;
}
// TODO: https://stackoverflow.com/questions/10463518/converting-em-to-px-in-javascript-and-getting-default-font-size
function getVisibleEditorLineHeight() {
    return $("#booksMenuButton").first().height();
}
function newDiv(parent, cls) {
    var div = $(document.createElement("div"));
    if (parent)
        div.appendTo(parent);
    if (cls)
        div.addClass(cls);
    return div;
}
/////
var MAX_ERRORS = 200;
var SourceEditor = /** @class */ (function () {
    function SourceEditor(path, mode) {
        this.dirtylisting = true;
        this.errormsgs = [];
        this.errorwidgets = [];
        this.path = path;
        this.mode = mode;
    }
    SourceEditor.prototype.createDiv = function (parent, text) {
        var div = document.createElement('div');
        div.setAttribute("class", "editor");
        parent.appendChild(div);
        var asmOverride = text && this.mode == 'verilog' && /__asm\b([\s\S]+?)\b__endasm\b/.test(text);
        this.newEditor(div, asmOverride);
        if (text) {
            this.setText(text); // TODO: this calls setCode() and builds... it shouldn't
        }
        this.setupEditor();
        return div;
    };
    SourceEditor.prototype.newEditor = function (parent, isAsmOverride) {
        var isAsm = isAsmOverride || this.mode == '6502' || this.mode == 'z80' || this.mode == 'jsasm' || this.mode == 'gas'; // TODO
        var lineWrap = this.mode == 'markdown';
        this.editor = CodeMirror(parent, {
            theme: 'mbo',
            lineNumbers: true,
            matchBrackets: true,
            tabSize: 8,
            indentAuto: true,
            lineWrapping: lineWrap,
            gutters: isAsm ? ["CodeMirror-linenumbers", "gutter-offset", "gutter-bytes", "gutter-clock", "gutter-info"]
                : ["CodeMirror-linenumbers", "gutter-offset", "gutter-info"],
        });
    };
    SourceEditor.prototype.setupEditor = function () {
        var _this = this;
        var timer;
        this.editor.on('changes', function (ed, changeobj) {
            clearTimeout(timer);
            timer = setTimeout(function () {
                ui_1.current_project.updateFile(_this.path, _this.editor.getValue());
            }, 300);
        });
        this.editor.on('cursorActivity', function (ed) {
            var start = _this.editor.getCursor(true);
            var end = _this.editor.getCursor(false);
            if (start.line == end.line && start.ch < end.ch && end.ch - start.ch < 80) {
                var name = _this.editor.getSelection();
                _this.inspect(name);
            }
            else {
                _this.inspect(null);
            }
        });
        //scrollProfileView(editor);
        this.editor.setOption("mode", this.mode);
    };
    SourceEditor.prototype.inspect = function (ident) {
        var result;
        if (ui_1.platform.inspect) {
            result = ui_1.platform.inspect(ident);
        }
        if (this.inspectWidget) {
            this.inspectWidget.clear();
            this.inspectWidget = null;
        }
        if (result) {
            var infospan = createTextSpan(result, "tooltipinfoline");
            var line = this.editor.getCursor().line;
            this.inspectWidget = this.editor.addLineWidget(line, infospan, { above: false });
        }
    };
    SourceEditor.prototype.setText = function (text) {
        var i, j;
        var oldtext = this.editor.getValue();
        if (oldtext != text) {
            // find minimum range to undo
            for (i = 0; i < oldtext.length && i < text.length && text[i] == oldtext[i]; i++) { }
            for (j = 0; j < oldtext.length && j < text.length && text[text.length - 1 - j] == oldtext[oldtext.length - 1 - j]; j++) { }
            //console.log(i,j,oldtext.substring(i,oldtext.length-j));
            this.replaceSelection(i, oldtext.length - j, text.substring(i, text.length - j)); // calls setCode()
            // clear history if setting empty editor
            if (oldtext == '') {
                this.editor.clearHistory();
            }
        }
    };
    SourceEditor.prototype.insertText = function (text) {
        var cur = this.editor.getCursor();
        this.editor.replaceRange(text, cur, cur);
    };
    SourceEditor.prototype.replaceSelection = function (start, end, text) {
        this.editor.setSelection(this.editor.posFromIndex(start), this.editor.posFromIndex(end));
        this.editor.replaceSelection(text);
    };
    SourceEditor.prototype.getValue = function () {
        return this.editor.getValue();
    };
    SourceEditor.prototype.getPath = function () { return this.path; };
    SourceEditor.prototype.addErrorMarker = function (line, msg) {
        var _this = this;
        var div = document.createElement("div");
        div.setAttribute("class", "tooltipbox tooltiperror");
        div.appendChild(document.createTextNode("\u24cd"));
        this.editor.setGutterMarker(line, "gutter-info", div);
        this.errormsgs.push({ line: line, msg: msg });
        // expand line widgets when mousing over errors
        $(div).mouseover(function (e) {
            _this.expandErrors();
        });
    };
    SourceEditor.prototype.addErrorLine = function (line, msg) {
        var errspan = createTextSpan(msg, "tooltiperrorline");
        this.errorwidgets.push(this.editor.addLineWidget(line, errspan));
    };
    SourceEditor.prototype.expandErrors = function () {
        var e;
        while (e = this.errormsgs.shift()) {
            this.addErrorLine(e.line, e.msg);
        }
    };
    SourceEditor.prototype.markErrors = function (errors) {
        // TODO: move cursor to error line if offscreen?
        this.clearErrors();
        var numLines = this.editor.lineCount();
        errors = errors.slice(0, MAX_ERRORS);
        for (var _i = 0, errors_1 = errors; _i < errors_1.length; _i++) {
            var info = errors_1[_i];
            // only mark errors with this filename, or without any filename
            if (!info.path || this.path.endsWith(info.path)) {
                var line = info.line - 1;
                if (line < 0 || line >= numLines)
                    line = 0;
                this.addErrorMarker(line, info.msg);
            }
        }
    };
    SourceEditor.prototype.clearErrors = function () {
        this.editor.clearGutter("gutter-info");
        this.refreshDebugState(false);
        this.dirtylisting = true;
        // clear line widgets
        this.errormsgs = [];
        while (this.errorwidgets.length)
            this.errorwidgets.shift().clear();
    };
    SourceEditor.prototype.getSourceFile = function () { return this.sourcefile; };
    SourceEditor.prototype.updateListing = function () {
        // update editor annotations
        // TODO: recreate editor if gutter-bytes is used (verilog)
        this.editor.clearGutter("gutter-info");
        this.editor.clearGutter("gutter-bytes");
        this.editor.clearGutter("gutter-offset");
        this.editor.clearGutter("gutter-clock");
        var lstlines = this.sourcefile.lines || [];
        for (var _i = 0, lstlines_1 = lstlines; _i < lstlines_1.length; _i++) {
            var info = lstlines_1[_i];
            if (info.offset >= 0) {
                this.setGutter("gutter-offset", info.line - 1, util_1.hex(info.offset & 0xffff, 4));
            }
            if (info.insns) {
                var insnstr = info.insns.length > 9 ? ("...") : info.insns;
                this.setGutter("gutter-bytes", info.line - 1, insnstr);
                if (info.iscode) {
                    // TODO: labels trick this part?
                    if (info.cycles) {
                        this.setGutter("gutter-clock", info.line - 1, info.cycles + "");
                    }
                    else if (ui_1.platform.getOpcodeMetadata) {
                        var opcode = parseInt(info.insns.split(" ")[0], 16);
                        var meta = ui_1.platform.getOpcodeMetadata(opcode, info.offset);
                        var clockstr = meta.minCycles + "";
                        this.setGutter("gutter-clock", info.line - 1, clockstr);
                    }
                }
            }
        }
    };
    SourceEditor.prototype.setGutter = function (type, line, text) {
        var lineinfo = this.editor.lineInfo(line);
        if (lineinfo && lineinfo.gutterMarkers && lineinfo.gutterMarkers[type]) {
            // do not replace existing marker
        }
        else {
            var textel = document.createTextNode(text);
            this.editor.setGutterMarker(line, type, textel);
        }
    };
    SourceEditor.prototype.setGutterBytes = function (line, s) {
        this.setGutter("gutter-bytes", line - 1, s);
    };
    SourceEditor.prototype.setTimingResult = function (result) {
        this.editor.clearGutter("gutter-bytes");
        // show the lines
        for (var _i = 0, _a = Object.keys(this.sourcefile.line2offset); _i < _a.length; _i++) {
            var line = _a[_i];
            var pc = this.sourcefile.line2offset[line];
            var minclocks = result.pc2minclocks[pc];
            var maxclocks = result.pc2maxclocks[pc];
            if (minclocks >= 0 && maxclocks >= 0) {
                var s;
                if (maxclocks == minclocks)
                    s = minclocks + "";
                else
                    s = minclocks + "-" + maxclocks;
                if (maxclocks == result.MAX_CLOCKS)
                    s += "+";
                this.setGutterBytes(parseInt(line), s);
            }
        }
    };
    SourceEditor.prototype.setCurrentLine = function (line, moveCursor) {
        var _this = this;
        var addCurrentMarker = function (line) {
            var div = document.createElement("div");
            div.style.color = '#66ffff';
            div.appendChild(document.createTextNode("\u25b6"));
            _this.editor.setGutterMarker(line, "gutter-info", div);
        };
        this.clearCurrentLine();
        if (line > 0) {
            addCurrentMarker(line - 1);
            if (moveCursor)
                this.editor.setSelection({ line: line, ch: 0 }, { line: line - 1, ch: 0 }, { scroll: true });
            this.currentDebugLine = line;
        }
    };
    SourceEditor.prototype.clearCurrentLine = function () {
        if (this.currentDebugLine) {
            this.editor.clearGutter("gutter-info");
            this.editor.setSelection(this.editor.getCursor());
            this.currentDebugLine = 0;
        }
    };
    SourceEditor.prototype.getActiveLine = function () {
        var state = ui_1.lastDebugState;
        if (state && state.c && this.sourcefile) {
            var EPC = state.c.EPC || state.c.PC;
            var line = this.sourcefile.findLineForOffset(EPC, 15);
            return line;
        }
        else
            return -1;
    };
    SourceEditor.prototype.refreshDebugState = function (moveCursor) {
        this.clearCurrentLine();
        var line = this.getActiveLine();
        if (line >= 0) {
            this.setCurrentLine(line, moveCursor);
            // TODO: switch to disasm?
        }
    };
    SourceEditor.prototype.refreshListing = function () {
        // lookup corresponding sourcefile for this file, using listing
        var lst = ui_1.current_project.getListingForFile(this.path);
        if (lst && lst.sourcefile && lst.sourcefile !== this.sourcefile) {
            this.sourcefile = lst.sourcefile;
            this.dirtylisting = true;
        }
        if (!this.sourcefile || !this.dirtylisting)
            return;
        this.dirtylisting = false;
        this.updateListing();
    };
    SourceEditor.prototype.refresh = function (moveCursor) {
        this.refreshListing();
        this.refreshDebugState(moveCursor);
    };
    SourceEditor.prototype.getLine = function (line) {
        return this.editor.getLine(line - 1);
    };
    SourceEditor.prototype.getCurrentLine = function () {
        return this.editor.getCursor().line + 1;
    };
    SourceEditor.prototype.getCursorPC = function () {
        var line = this.getCurrentLine();
        while (this.sourcefile && line >= 0) {
            var pc = this.sourcefile.line2offset[line];
            if (pc >= 0)
                return pc;
            line--;
        }
        return -1;
    };
    SourceEditor.prototype.undoStep = function () {
        this.editor.execCommand('undo');
    };
    return SourceEditor;
}());
exports.SourceEditor = SourceEditor;
///
var disasmWindow = 1024; // disassemble this many bytes around cursor
var DisassemblerView = /** @class */ (function () {
    function DisassemblerView() {
    }
    DisassemblerView.prototype.getDisasmView = function () { return this.disasmview; };
    DisassemblerView.prototype.createDiv = function (parent) {
        var div = document.createElement('div');
        div.setAttribute("class", "editor");
        parent.appendChild(div);
        this.newEditor(div);
        return div;
    };
    DisassemblerView.prototype.newEditor = function (parent) {
        this.disasmview = CodeMirror(parent, {
            mode: 'z80',
            theme: 'cobalt',
            tabSize: 8,
            readOnly: true,
            styleActiveLine: true
        });
    };
    // TODO: too many globals
    DisassemblerView.prototype.refresh = function (moveCursor) {
        var state = ui_1.lastDebugState || ui_1.platform.saveState(); // TODO?
        var pc = state.c ? state.c.PC : 0;
        var curline = 0;
        var selline = 0;
        var addr2symbol = (ui_1.platform.debugSymbols && ui_1.platform.debugSymbols.addr2symbol) || {};
        // TODO: not perfect disassembler
        var disassemble = function (start, end) {
            if (start < 0)
                start = 0;
            if (end > 0xffff)
                end = 0xffff;
            // TODO: use pc2visits
            var a = start;
            var s = "";
            while (a < end) {
                var disasm = ui_1.platform.disassemble(a, ui_1.platform.readAddress.bind(ui_1.platform));
                /* TODO: look thru all source files
                var srclinenum = sourcefile && this.sourcefile.offset2line[a];
                if (srclinenum) {
                  var srcline = getActiveEditor().getLine(srclinenum);
                  if (srcline && srcline.trim().length) {
                    s += "; " + srclinenum + ":\t" + srcline + "\n";
                    curline++;
                  }
                }
                */
                var bytes = "";
                var comment = "";
                for (var i = 0; i < disasm.nbytes; i++)
                    bytes += util_1.hex(ui_1.platform.readAddress(a + i));
                while (bytes.length < 14)
                    bytes += ' ';
                var dstr = disasm.line;
                if (addr2symbol && disasm.isaddr) {
                    dstr = dstr.replace(/([^#])[$]([0-9A-F]+)/, function (substr) {
                        var args = [];
                        for (var _i = 1; _i < arguments.length; _i++) {
                            args[_i - 1] = arguments[_i];
                        }
                        var addr = parseInt(args[1], 16);
                        var sym = addr2symbol[addr];
                        if (sym)
                            return (args[0] + sym);
                        sym = addr2symbol[addr - 1];
                        if (sym)
                            return (args[0] + sym + "+1");
                        return substr;
                    });
                }
                if (addr2symbol) {
                    var sym = addr2symbol[a];
                    if (sym) {
                        comment = "; " + sym;
                    }
                }
                var dline = util_1.hex(parseInt(a), 4) + "\t" + util_1.rpad(bytes, 14) + "\t" + util_1.rpad(dstr, 30) + comment + "\n";
                s += dline;
                if (a == pc)
                    selline = curline;
                curline++;
                a += disasm.nbytes || 1;
            }
            return s;
        };
        var text = disassemble(pc - disasmWindow, pc) + disassemble(pc, pc + disasmWindow);
        this.disasmview.setValue(text);
        if (moveCursor) {
            this.disasmview.setCursor(selline, 0);
            jumpToLine(this.disasmview, selline);
        }
    };
    DisassemblerView.prototype.getCursorPC = function () {
        var line = this.disasmview.getCursor().line;
        if (line >= 0) {
            var toks = this.disasmview.getLine(line).trim().split(/\s+/);
            if (toks && toks.length >= 1) {
                var pc = parseInt(toks[0], 16);
                if (pc >= 0)
                    return pc;
            }
        }
        return -1;
    };
    return DisassemblerView;
}());
exports.DisassemblerView = DisassemblerView;
///
var ListingView = /** @class */ (function (_super) {
    __extends(ListingView, _super);
    function ListingView(lstfn) {
        var _this = _super.call(this) || this;
        _this.path = lstfn;
        return _this;
    }
    ListingView.prototype.refreshListing = function () {
        // lookup corresponding assemblyfile for this file, using listing
        var lst = ui_1.current_project.getListingForFile(this.path);
        // TODO?
        if (lst && lst.assemblyfile) {
            this.assemblyfile = lst.assemblyfile;
        }
        else if (lst && lst.sourcefile) {
            this.assemblyfile = lst.sourcefile;
        }
    };
    ListingView.prototype.refresh = function (moveCursor) {
        this.refreshListing();
        if (!this.assemblyfile)
            return; // TODO?
        var state = ui_1.lastDebugState || ui_1.platform.saveState(); // TODO?
        var pc = state.c ? (state.c.EPC || state.c.PC) : 0;
        var asmtext = this.assemblyfile.text;
        var disasmview = this.getDisasmView();
        disasmview.setValue(asmtext);
        var debugging = true; // TODO: platform.isDebugging && platform.isDebugging();
        var findPC = debugging ? pc : -1;
        if (findPC >= 0 && this.assemblyfile) {
            var lineno = this.assemblyfile.findLineForOffset(findPC, 15);
            if (lineno && moveCursor) {
                // set cursor while debugging
                if (debugging)
                    disasmview.setCursor(lineno - 1, 0);
                jumpToLine(disasmview, lineno - 1);
            }
        }
    };
    return ListingView;
}(DisassemblerView));
exports.ListingView = ListingView;
///
// TODO: make it use debug state
// TODO: make it safe (load/restore state?)
var MemoryView = /** @class */ (function () {
    function MemoryView() {
        this.recreateOnResize = true;
    }
    MemoryView.prototype.createDiv = function (parent) {
        var div = document.createElement('div');
        div.setAttribute("class", "memdump");
        parent.appendChild(div);
        this.showMemoryWindow(parent, div);
        return this.maindiv = div;
    };
    MemoryView.prototype.showMemoryWindow = function (workspace, parent) {
        var _this = this;
        this.memorylist = new VirtualList({
            w: $(workspace).width(),
            h: $(workspace).height(),
            itemHeight: getVisibleEditorLineHeight(),
            totalRows: 0x2000,
            generatorFn: function (row) {
                var s = _this.getMemoryLineAt(row);
                var linediv = document.createElement("div");
                if (_this.dumplines) {
                    var dlr = _this.dumplines[row];
                    if (dlr)
                        linediv.classList.add('seg_' + _this.getMemorySegment(_this.dumplines[row].a));
                }
                linediv.appendChild(document.createTextNode(s));
                return linediv;
            }
        });
        $(parent).append(this.memorylist.container);
        this.tick();
        if (ui_1.compparams && this.dumplines)
            this.scrollToAddress(ui_1.compparams.data_start);
    };
    MemoryView.prototype.scrollToAddress = function (addr) {
        this.memorylist.scrollToItem(this.findMemoryWindowLine(addr));
    };
    MemoryView.prototype.refresh = function () {
        this.dumplines = null;
        this.tick();
    };
    MemoryView.prototype.tick = function () {
        var _this = this;
        if (this.memorylist) {
            $(this.maindiv).find('[data-index]').each(function (i, e) {
                var div = $(e);
                var row = parseInt(div.attr('data-index'));
                var oldtext = div.text();
                var newtext = _this.getMemoryLineAt(row);
                if (oldtext != newtext)
                    div.text(newtext);
            });
        }
    };
    MemoryView.prototype.getMemoryLineAt = function (row) {
        var offset = row * 16;
        var n1 = 0;
        var n2 = 16;
        var sym;
        if (this.getDumpLines()) {
            var dl = this.dumplines[row];
            if (dl) {
                offset = dl.a & 0xfff0;
                n1 = dl.a - offset;
                n2 = n1 + dl.l;
                sym = dl.s;
            }
            else {
                return '.';
            }
        }
        var s = util_1.hex(offset + n1, 4) + ' ';
        for (var i = 0; i < n1; i++)
            s += '   ';
        if (n1 > 8)
            s += ' ';
        for (var i = n1; i < n2; i++) {
            var read = this.readAddress(offset + i);
            if (i == 8)
                s += ' ';
            s += ' ' + (typeof read == 'number' ? util_1.hex(read, 2) : '??');
        }
        for (var i = n2; i < 16; i++)
            s += '   ';
        if (sym)
            s += '  ' + sym;
        return s;
    };
    MemoryView.prototype.readAddress = function (n) {
        return ui_1.platform.readAddress(n);
    };
    MemoryView.prototype.getDumpLineAt = function (line) {
        var d = this.dumplines[line];
        if (d) {
            return d.a + " " + d.s;
        }
    };
    // TODO: addr2symbol for ca65; and make it work without symbols
    MemoryView.prototype.getDumpLines = function () {
        var addr2sym = (ui_1.platform.debugSymbols && ui_1.platform.debugSymbols.addr2symbol) || {};
        if (!this.dumplines) {
            this.dumplines = [];
            var ofs = 0;
            var sym;
            for (var _i = 0, _a = Object.keys(addr2sym); _i < _a.length; _i++) {
                var _nextofs = _a[_i];
                var nextofs = parseInt(_nextofs); // convert from string (stupid JS)
                var nextsym = addr2sym[nextofs];
                if (sym) {
                    // ignore certain symbols
                    if (sym.endsWith('_SIZE__') || sym.endsWith('_LAST__') || sym.endsWith('STACKSIZE__') || sym.endsWith('FILEOFFS__') || sym.startsWith('l__'))
                        sym = '';
                    if (MemoryView.IGNORE_SYMS[sym]) {
                        ofs = nextofs;
                    }
                    else {
                        while (ofs < nextofs) {
                            var ofs2 = (ofs + 16) & 0xffff0;
                            if (ofs2 > nextofs)
                                ofs2 = nextofs;
                            //if (ofs < 1000) console.log(ofs, ofs2, nextofs, sym);
                            this.dumplines.push({ a: ofs, l: ofs2 - ofs, s: sym });
                            ofs = ofs2;
                        }
                    }
                }
                sym = nextsym;
            }
        }
        return this.dumplines;
    };
    // TODO: use segments list?
    MemoryView.prototype.getMemorySegment = function (a) {
        if (ui_1.compparams) {
            if (a >= ui_1.compparams.data_start && a < ui_1.compparams.data_start + ui_1.compparams.data_size) {
                if (ui_1.platform.getSP && a >= ui_1.platform.getSP() - 15)
                    return 'stack';
                else
                    return 'data';
            }
            else if (a >= ui_1.compparams.code_start && a < ui_1.compparams.code_start + (ui_1.compparams.code_size || ui_1.compparams.rom_size))
                return 'code';
        }
        var segments = ui_1.current_project.segments;
        if (segments) {
            for (var _i = 0, segments_1 = segments; _i < segments_1.length; _i++) {
                var seg = segments_1[_i];
                if (a >= seg.start && a < seg.start + seg.size) {
                    if (seg.type == 'rom')
                        return 'code';
                    if (seg.type == 'ram')
                        return 'data';
                    if (seg.type == 'io')
                        return 'io';
                }
            }
        }
        return 'unknown';
    };
    MemoryView.prototype.findMemoryWindowLine = function (a) {
        for (var i = 0; i < this.dumplines.length; i++)
            if (this.dumplines[i].a >= a)
                return i;
    };
    MemoryView.IGNORE_SYMS = { s__INITIALIZER: true, /* s__GSINIT:true, */ _color_prom: true };
    return MemoryView;
}());
exports.MemoryView = MemoryView;
var VRAMMemoryView = /** @class */ (function (_super) {
    __extends(VRAMMemoryView, _super);
    function VRAMMemoryView() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    VRAMMemoryView.prototype.readAddress = function (n) {
        return ui_1.platform.readVRAMAddress(n);
    };
    VRAMMemoryView.prototype.getMemorySegment = function (a) {
        return 'video';
    };
    VRAMMemoryView.prototype.getDumpLines = function () {
        return null;
    };
    return VRAMMemoryView;
}(MemoryView));
exports.VRAMMemoryView = VRAMMemoryView;
///
var BinaryFileView = /** @class */ (function () {
    function BinaryFileView(path, data) {
        this.recreateOnResize = true;
        this.path = path;
        this.data = data;
    }
    BinaryFileView.prototype.createDiv = function (parent) {
        var div = document.createElement('div');
        div.setAttribute("class", "memdump");
        parent.appendChild(div);
        this.showMemoryWindow(parent, div);
        return this.maindiv = div;
    };
    BinaryFileView.prototype.showMemoryWindow = function (workspace, parent) {
        var _this = this;
        this.memorylist = new VirtualList({
            w: $(workspace).width(),
            h: $(workspace).height(),
            itemHeight: getVisibleEditorLineHeight(),
            totalRows: ((this.data.length + 15) >> 4),
            generatorFn: function (row) {
                var s = _this.getMemoryLineAt(row);
                var linediv = document.createElement("div");
                linediv.appendChild(document.createTextNode(s));
                return linediv;
            }
        });
        $(parent).append(this.memorylist.container);
    };
    BinaryFileView.prototype.getMemoryLineAt = function (row) {
        var offset = row * 16;
        var n1 = 0;
        var n2 = 16;
        var s = util_1.hex(offset + n1, 4) + ' ';
        for (var i = 0; i < n1; i++)
            s += '   ';
        if (n1 > 8)
            s += ' ';
        for (var i = n1; i < n2; i++) {
            var read = this.data[offset + i];
            if (i == 8)
                s += ' ';
            s += ' ' + (read >= 0 ? util_1.hex(read, 2) : '  ');
        }
        return s;
    };
    BinaryFileView.prototype.refresh = function () {
    };
    BinaryFileView.prototype.getPath = function () { return this.path; };
    return BinaryFileView;
}());
exports.BinaryFileView = BinaryFileView;
///
var MemoryMapView = /** @class */ (function () {
    function MemoryMapView() {
    }
    MemoryMapView.prototype.createDiv = function (parent) {
        this.maindiv = newDiv(parent, 'vertical-scroll');
        return this.maindiv[0];
    };
    // TODO: overlapping segments (e.g. ROM + LC)
    MemoryMapView.prototype.addSegment = function (seg) {
        var offset = $('<div class="col-md-1 segment-offset"/>');
        offset.text('$' + util_1.hex(seg.start, 4));
        var segdiv = $('<div class="col-md-4 segment"/>');
        if (seg.last)
            segdiv.text(seg.name + " (" + (seg.last - seg.start) + " / " + seg.size + " bytes used)");
        else
            segdiv.text(seg.name + " (" + seg.size + " bytes)");
        if (seg.size >= 256) {
            var pad = (Math.log(seg.size) - Math.log(256)) * 0.5;
            segdiv.css('padding-top', pad + 'em');
            segdiv.css('padding-bottom', pad + 'em');
        }
        if (seg.type) {
            segdiv.addClass('segment-' + seg.type);
        }
        var row = $('<div class="row"/>').append(offset, segdiv);
        var container = $('<div class="container"/>').append(row);
        this.maindiv.append(container);
        segdiv.click(function () {
            var memview = ui_1.projectWindows.createOrShow('#memory');
            memview.scrollToAddress(seg.start);
            // TODO: this doesn't update nav bar
        });
    };
    MemoryMapView.prototype.refresh = function () {
        this.maindiv.empty();
        var segments = ui_1.current_project.segments;
        if (segments) {
            var curofs = 0;
            for (var _i = 0, segments_2 = segments; _i < segments_2.length; _i++) {
                var seg = segments_2[_i];
                //var used = seg.last ? (seg.last-seg.start) : seg.size;
                if (seg.start > curofs)
                    this.addSegment({ name: '', start: curofs, size: seg.start - curofs });
                this.addSegment(seg);
                curofs = seg.start + seg.size;
            }
        }
    };
    return MemoryMapView;
}());
exports.MemoryMapView = MemoryMapView;
///
var ProfileView = /** @class */ (function () {
    function ProfileView() {
        this.symcache = new Map();
        this.recreateOnResize = true;
    }
    ProfileView.prototype.createDiv = function (parent) {
        var div = document.createElement('div');
        div.setAttribute("class", "profiler");
        parent.appendChild(div);
        this.showMemoryWindow(parent, div);
        return this.maindiv = div;
    };
    ProfileView.prototype.showMemoryWindow = function (workspace, parent) {
        var _this = this;
        this.profilelist = new VirtualList({
            w: $(workspace).width(),
            h: $(workspace).height(),
            itemHeight: getVisibleEditorLineHeight(),
            totalRows: 262,
            generatorFn: function (row) {
                var linediv = document.createElement("div");
                _this.addProfileLine(linediv, row);
                return linediv;
            }
        });
        $(parent).append(this.profilelist.container);
        this.symcache = new Map();
        this.refresh();
    };
    ProfileView.prototype.addProfileLine = function (div, row) {
        div.appendChild(createTextSpan(util_1.lpad(row + ':', 4), "profiler-lineno"));
        if (!this.out)
            return;
        var f = this.out.frame;
        if (!f)
            return;
        var l = f.lines[row];
        if (!l)
            return;
        var lastsym = '';
        var canDebug = ui_1.platform.runToFrameClock;
        var _loop_1 = function (i) {
            var pc = f.iptab[i];
            var sym = this_1.symcache[pc];
            if (!sym) {
                sym = baseplatform_1.lookupSymbol(ui_1.platform, pc, false);
                this_1.symcache[pc] = sym;
            }
            if (sym != lastsym) {
                cls = "profiler";
                if (sym.startsWith('_'))
                    cls = "profiler-cident";
                else if (sym.startsWith('@'))
                    cls = "profiler-local";
                else if (/^\d*[.]/.exec(sym))
                    cls = "profiler-local";
                span = createTextSpan(' ' + sym, cls);
                if (canDebug) {
                    $(span).click(function () {
                        ui_1.platform.runToFrameClock(i);
                    });
                }
                div.appendChild(span);
                lastsym = sym;
            }
        };
        var this_1 = this, cls, span;
        for (var i = l.start; i <= l.end; i++) {
            _loop_1(i);
        }
    };
    ProfileView.prototype.refresh = function () {
        this.tick();
        this.symcache.clear();
    };
    ProfileView.prototype.tick = function () {
        var _this = this;
        if (this.profilelist) {
            $(this.maindiv).find('[data-index]').each(function (i, e) {
                var div = $(e);
                var row = parseInt(div.attr('data-index'));
                div.empty();
                _this.addProfileLine(div[0], row);
            });
        }
    };
    ProfileView.prototype.setVisible = function (showing) {
        if (!this.prof) {
            this.prof = new recorder_1.EmuProfilerImpl(ui_1.platform);
        }
        if (showing)
            this.out = this.prof.start();
        else
            this.prof.stop();
    };
    return ProfileView;
}());
exports.ProfileView = ProfileView;
///
var AssetEditorView = /** @class */ (function () {
    function AssetEditorView() {
    }
    AssetEditorView.prototype.createDiv = function (parent) {
        this.maindiv = newDiv(parent, "vertical-scroll");
        return this.maindiv[0];
    };
    AssetEditorView.prototype.clearAssets = function () {
        this.rootnodes = [];
        this.deferrednodes = [];
    };
    AssetEditorView.prototype.registerAsset = function (type, node, deferred) {
        this.rootnodes.push(node);
        if (deferred) {
            if (deferred > 1)
                this.deferrednodes.push(node);
            else
                this.deferrednodes.unshift(node);
        }
        else {
            node.refreshRight();
        }
    };
    AssetEditorView.prototype.getPalettes = function (matchlen) {
        var result = [];
        this.rootnodes.forEach(function (node) {
            while (node != null) {
                if (node instanceof pixed.PaletteFormatToRGB) {
                    // TODO: move to node class?
                    var palette = node.palette;
                    // match full palette length?
                    if (matchlen == palette.length) {
                        result.push({ node: node, name: "Palette", palette: palette });
                    }
                    // look at palette slices
                    if (node.layout) {
                        node.layout.forEach(function (_a) {
                            var name = _a[0], start = _a[1], len = _a[2];
                            if (start < palette.length) {
                                if (len == matchlen) {
                                    var rgbs = palette.slice(start, start + len);
                                    result.push({ node: node, name: name, palette: rgbs });
                                }
                                else if (-len == matchlen) { // reverse order
                                    var rgbs = palette.slice(start, start - len);
                                    rgbs.reverse();
                                    result.push({ node: node, name: name, palette: rgbs });
                                }
                                else if (len + 1 == matchlen) {
                                    var rgbs = new Uint32Array(matchlen);
                                    rgbs[0] = palette[0];
                                    rgbs.set(palette.slice(start, start + len), 1);
                                    result.push({ node: node, name: name, palette: rgbs });
                                }
                            }
                        });
                    }
                    break;
                }
                node = node.right;
            }
        });
        return result;
    };
    AssetEditorView.prototype.getTilemaps = function (matchlen) {
        var result = [];
        this.rootnodes.forEach(function (node) {
            while (node != null) {
                if (node instanceof pixed.Palettizer) {
                    var rgbimgs = node.rgbimgs;
                    if (rgbimgs && rgbimgs.length >= matchlen) {
                        result.push({ node: node, name: "Tilemap", images: node.images, rgbimgs: rgbimgs }); // TODO
                    }
                }
                node = node.right;
            }
        });
        return result;
    };
    AssetEditorView.prototype.isEditing = function () {
        return this.cureditordiv != null;
    };
    AssetEditorView.prototype.getCurrentEditNode = function () {
        return this.cureditnode;
    };
    AssetEditorView.prototype.setCurrentEditor = function (div, editing, node) {
        var timeout = 250;
        if (this.cureditordiv != div) {
            if (this.cureditordiv) {
                this.cureditordiv.hide(timeout);
                this.cureditordiv = null;
            }
            if (div) {
                this.cureditordiv = div;
                this.cureditordiv.show();
                this.cureditordiv[0].scrollIntoView({ behavior: "smooth", block: "center" });
                //setTimeout(() => { this.cureditordiv[0].scrollIntoView({behavior: "smooth", block: "center"}) }, timeout);
            }
        }
        if (this.cureditelem) {
            this.cureditelem.removeClass('selected');
            this.cureditelem = null;
        }
        if (editing) {
            this.cureditelem = editing;
            this.cureditelem.addClass('selected');
        }
        while (node.left) {
            node = node.left;
        }
        this.cureditnode = node;
    };
    AssetEditorView.prototype.scanFileTextForAssets = function (id, data) {
        // scan file for assets
        // /*{json}*/ or ;;{json};;
        // TODO: put before ident, look for = {
        var result = [];
        var re1 = /[/;][*;]([{].+[}])[*;][/;]/g;
        var m;
        while (m = re1.exec(data)) {
            var start = m.index + m[0].length;
            var end;
            // TODO: verilog end
            if (ui_1.platform_id == 'verilog') {
                end = data.indexOf("end", start); // asm
            }
            else if (m[0].startsWith(';;')) {
                end = data.indexOf(';;', start); // asm
            }
            else {
                end = data.indexOf(';', start); // C
            }
            //console.log(id, start, end, m[1], data.substring(start,end));
            if (end > start) {
                try {
                    var jsontxt = m[1].replace(/([A-Za-z]+):/g, '"$1":'); // fix lenient JSON
                    var json = JSON.parse(jsontxt);
                    // TODO: name?
                    result.push({ fileid: id, fmt: json, start: start, end: end });
                }
                catch (e) {
                    console.log(e);
                }
            }
        }
        // look for DEF_METASPRITE_2x2(playerRStand, 0xd8, 0)
        // TODO: could also look in ROM
        var re2 = /DEF_METASPRITE_(\d+)x(\d+)[(](\w+),\s*(\w+),\s*(\w+)/gi;
        while (m = re2.exec(data)) {
            var width = parseInt(m[1]);
            var height = parseInt(m[2]);
            var ident = m[3];
            var tile = parseInt(m[4]);
            var attr = parseInt(m[5]);
            var metadefs = [];
            for (var x = 0; x < width; x++) {
                for (var y = 0; y < height; y++) {
                    metadefs.push({ x: x * 8, y: y * 8, tile: tile, attr: attr });
                }
            }
            var meta = { defs: metadefs, width: width * 8, height: height * 8 };
            result.push({ fileid: id, label: ident, meta: meta });
        }
        return result;
    };
    // TODO: move to pixeleditor.ts?
    AssetEditorView.prototype.addPaletteEditorViews = function (parentdiv, pal2rgb, callback) {
        var _this = this;
        var adual = $('<div class="asset_dual"/>').appendTo(parentdiv);
        var aeditor = $('<div class="asset_editor"/>').hide(); // contains editor, when selected
        // TODO: they need to update when refreshed from right
        var allrgbimgs = [];
        pal2rgb.getAllColors().forEach(function (rgba) { allrgbimgs.push(new Uint32Array([rgba])); }); // array of array of 1 rgb color (for picker)
        var atable = $('<table/>').appendTo(adual);
        aeditor.appendTo(adual);
        // make default layout if not exists
        var layout = pal2rgb.layout;
        if (!layout) {
            var len = pal2rgb.palette.length;
            var imgsperline = len > 32 ? 8 : 4; // TODO: use 'n'?
            layout = [];
            for (var i = 0; i < len; i += imgsperline) {
                layout.push(["", i, Math.min(len - i, imgsperline)]);
            }
        }
        function updateCell(cell, j) {
            var val = pal2rgb.words[j];
            var rgb = pal2rgb.palette[j];
            var hexcol = '#' + util_1.hex(util_1.rgb2bgr(rgb), 6);
            var textcol = (rgb & 0x008000) ? 'black' : 'white';
            cell.text(util_1.hex(val, 2)).css('background-color', hexcol).css('color', textcol);
        }
        // iterate over each row of the layout
        layout.forEach(function (_a) {
            var name = _a[0], start = _a[1], len = _a[2];
            if (start < pal2rgb.palette.length) { // skip row if out of range
                var arow = $('<tr/>').appendTo(atable);
                $('<td/>').text(name).appendTo(arow);
                var inds = [];
                for (var k = start; k < start + Math.abs(len); k++)
                    inds.push(k);
                if (len < 0)
                    inds.reverse();
                inds.forEach(function (i) {
                    var cell = $('<td/>').addClass('asset_cell asset_editable').appendTo(arow);
                    updateCell(cell, i);
                    cell.click(function (e) {
                        var chooser = new pixed.ImageChooser();
                        chooser.rgbimgs = allrgbimgs;
                        chooser.width = 1;
                        chooser.height = 1;
                        chooser.recreate(aeditor, function (index, newvalue) {
                            callback(i, index);
                            updateCell(cell, i);
                        });
                        _this.setCurrentEditor(aeditor, cell, pal2rgb);
                    });
                });
            }
        });
    };
    AssetEditorView.prototype.addPixelEditor = function (parentdiv, firstnode, fmt) {
        // data -> pixels
        fmt.xform = 'scale(2)';
        var mapper = new pixed.Mapper(fmt);
        // TODO: rotate node?
        firstnode.addRight(mapper);
        // pixels -> RGBA
        var palizer = new pixed.Palettizer(this, fmt);
        mapper.addRight(palizer);
        // add view objects
        palizer.addRight(new pixed.CharmapEditor(this, newDiv(parentdiv), fmt));
    };
    AssetEditorView.prototype.addPaletteEditor = function (parentdiv, firstnode, palfmt) {
        // palette -> RGBA
        var pal2rgb = new pixed.PaletteFormatToRGB(palfmt);
        firstnode.addRight(pal2rgb);
        // TODO: refresh twice?
        firstnode.refreshRight();
        // TODO: add view objects
        // TODO: show which one is selected?
        this.addPaletteEditorViews(parentdiv, pal2rgb, function (index, newvalue) {
            console.log('set entry', index, '=', newvalue);
            // TODO: this forces update of palette rgb colors and file data
            firstnode.words[index] = newvalue;
            pal2rgb.words = null;
            pal2rgb.updateRight();
            pal2rgb.refreshLeft();
        });
    };
    AssetEditorView.prototype.ensureFileDiv = function (fileid) {
        var divid = this.getFileDivId(fileid);
        var body = $(document.getElementById(divid));
        if (body.length === 0) {
            var filediv = newDiv(this.maindiv, 'asset_file');
            var header = newDiv(filediv, 'asset_file_header').text(fileid);
            body = newDiv(filediv).attr('id', divid).addClass('disable-select');
        }
        return body;
    };
    AssetEditorView.prototype.refreshAssetsInFile = function (fileid, data) {
        var nassets = 0;
        // TODO: check fmt w/h/etc limits
        // TODO: defer editor creation
        // TODO: only refresh when needed
        if (ui_1.platform_id.startsWith('nes') && fileid.endsWith('.chr') && data instanceof Uint8Array) {
            // is this a NES CHR?
            var node = new pixed.FileDataNode(ui_1.projectWindows, fileid);
            var neschrfmt = { w: 8, h: 8, bpp: 1, count: (data.length >> 4), brev: true, np: 2, pofs: 8, remap: [0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12] }; // TODO
            this.addPixelEditor(this.ensureFileDiv(fileid), node, neschrfmt);
            this.registerAsset("charmap", node, 1);
            nassets++;
        }
        else if (ui_1.platform_id.startsWith('nes') && fileid.endsWith('.pal') && data instanceof Uint8Array) {
            // is this a NES PAL?
            var node = new pixed.FileDataNode(ui_1.projectWindows, fileid);
            var nespalfmt = { pal: "nes", layout: "nes" };
            this.addPaletteEditor(this.ensureFileDiv(fileid), node, nespalfmt);
            this.registerAsset("palette", node, 0);
            nassets++;
        }
        else if (typeof data === 'string') {
            var textfrags = this.scanFileTextForAssets(fileid, data);
            for (var _i = 0, textfrags_1 = textfrags; _i < textfrags_1.length; _i++) {
                var frag = textfrags_1[_i];
                if (frag.fmt) {
                    var label = fileid; // TODO: label
                    var node = new pixed.TextDataNode(ui_1.projectWindows, fileid, label, frag.start, frag.end);
                    var first = node;
                    // rle-compressed? TODO: how to edit?
                    if (frag.fmt.comp == 'rletag') {
                        node = node.addRight(new pixed.Compressor());
                    }
                    // is this a nes nametable?
                    if (frag.fmt.map == 'nesnt') {
                        node = node.addRight(new pixed.NESNametableConverter(this));
                        node = node.addRight(new pixed.Palettizer(this, { w: 8, h: 8, bpp: 4 }));
                        var fmt = { w: 8 * (frag.fmt.w || 32), h: 8 * (frag.fmt.h || 30), count: 1 }; // TODO: can't do custom sizes
                        node = node.addRight(new pixed.MapEditor(this, newDiv(this.ensureFileDiv(fileid)), fmt));
                        this.registerAsset("nametable", first, 2);
                        nassets++;
                    }
                    // is this a bitmap?
                    else if (frag.fmt.w > 0 && frag.fmt.h > 0) {
                        this.addPixelEditor(this.ensureFileDiv(fileid), node, frag.fmt);
                        this.registerAsset("charmap", first, 1);
                        nassets++;
                    }
                    // is this a palette?
                    else if (frag.fmt.pal) {
                        this.addPaletteEditor(this.ensureFileDiv(fileid), node, frag.fmt);
                        this.registerAsset("palette", first, 0);
                        nassets++;
                    }
                    else {
                        // TODO: other kinds of resources?
                    }
                }
            }
        }
        return nassets;
    };
    AssetEditorView.prototype.getFileDivId = function (id) {
        return '__asset__' + util_1.safeident(id);
    };
    // TODO: recreate editors when refreshing
    // TODO: look for changes, not moveCursor
    AssetEditorView.prototype.refresh = function (moveCursor) {
        var _this = this;
        // clear and refresh all files/nodes?
        if (moveCursor) {
            this.maindiv.empty();
            this.clearAssets();
            ui_1.current_project.iterateFiles(function (fileid, data) {
                try {
                    var nassets = _this.refreshAssetsInFile(fileid, data);
                }
                catch (e) {
                    console.log(e);
                    _this.ensureFileDiv(fileid).text(e + ""); // TODO: error msg?
                }
            });
            console.log("Found " + this.rootnodes.length + " assets");
            this.deferrednodes.forEach(function (node) {
                try {
                    node.refreshRight();
                }
                catch (e) {
                    console.log(e);
                    alert(e + "");
                }
            });
            this.deferrednodes = [];
        }
        else {
            // only refresh nodes if not actively editing
            // since we could be in the middle of an operation that hasn't been committed
            for (var _i = 0, _a = this.rootnodes; _i < _a.length; _i++) {
                var node = _a[_i];
                if (node !== this.getCurrentEditNode()) {
                    node.refreshRight();
                }
            }
        }
    };
    AssetEditorView.prototype.setVisible = function (showing) {
        // TODO: make into toolbar?
        if (showing) {
            Mousetrap.bind('ctrl+z', ui_1.projectWindows.undoStep.bind(ui_1.projectWindows));
        }
        else {
            Mousetrap.unbind('ctrl+z');
        }
    };
    return AssetEditorView;
}());
exports.AssetEditorView = AssetEditorView;
//# sourceMappingURL=views.js.map